#include <math.h>
#include "sphrspwa.h"

vector vecttimes(float a, vector b)
{
  b.x *= a ;
  b.y *= a ;
  b.z *= a ;
  return b ;
}

vector vectsum(vector a, vector b)
{
  a.x += b.x ;
  a.y += b.y ;
  a.z += b.z ;
  return a ;
}

vector vectdiff(vector a, vector b)
{
  a.x -= b.x ;
  a.y -= b.y ;
  a.z -= b.z ;
  return a ;
}

float vectdot(vector a, vector b)
{
  a.x  = a.x*b.x;
  a.x += a.y*b.y;
  a.x += a.z*b.z;
  return a.x ;
}

vector vectcross(vector a, vector b)
{
  vector c ;
  c.x = a.y * b.z - b.y * a.z ;
  c.y = a.z * b.x - b.z * a.x ;
  c.z = a.x * b.y - b.x * a.y ;
  return c ;
}

float vabs2(vector a)
{
  return vectdot(a,a);
}

float vabs(vector a)
{
  return sqrt(vectdot(a,a));
}

vector unitvect(vector b)
{
  float flt;
  flt = vabs(b);
  if (flt!=0.0)
    return vecttimes(1.0/flt,b);
  else
    return vecttimes(0.0,b);
}

void eraseplayernsrte(float lon, float lat, float anglook, float prad, float v, float angtra)
// args: lon, lat, bearing of look, saucer radius, speed, angle of travel
// like drawplayernsr erasing but attempting to just erase the trailing edge
{
  float theta, rp; //parameters swept to draw the saucer
  float lonp, latp;  //lon,lat parameters, functions of theta, rp
  int x,y;  // screen xy of saucer point
  float rrorr;
  float prada,prada2;
  float thetasalt;
  float coslatp;
  float rc, rcb,rcc, rcd, rc2;
  int erase;
  float radialfudge=0.0;
  prada=prad*M_PI/Y_RES; // saucer radius in pixels converted to angle
  if (v==0.0) return;
  //  if(v<vinc*8)v*=8.04;else v+=vinc; // speed fudge
  v+=vinc*6; // speed fudge
  radialfudge = prada*.075;
  rcc = v*v-prada*prada ;
  //saucer
  rp=0.0; //
  while (rp<=prada)
  {
    thetasalt = ( (int)( (lon + lat + 60*rp)*400 )%36 )*M_PI/18.0; // salt from drawplayernsr
    theta=thetasalt;
    while (theta<(M_PIT2+thetasalt))
    {
      erase=0;
      // rc = rcb +- sqrt( rcb*rcb - rcc )
      rcb = v * cos(theta-angtra) ;
      rcd = rcb*rcb - rcc ;
      if(rcd < 0.0)
        erase=1;
      else
      {
        rcd = sqrt(rcd) ;
        rc  = rcb + rcd ;
        rc2 = rcb - rcd ;
        if ( rc*rc2<0.0 )
        { //it's always this case, rc*rc2<0.0, except for extremely high speed
          //if (cos(theta-angtra)<0.5)
          if ( rp+radialfudge >= (max(rc,rc2)) ) erase=1;
        }
        else
        if ( rc>0.0 && rc2>0.0 )
        {
          if ( rp-radialfudge <= (min(rc,rc2)) ) erase=1;
        }
        else
        if ( rc<0.0 && rc2<0.0 )
        {
          erase=1;
        }
        else
        if ( rc*rc2==0.0 )
        {
          if (rc+rc2<=0.0) erase=1;
        }
      }
      if (erase)
      {
        surfacerun(lon, lat, theta, rp, &lonp, &latp);
        projec(lonp,latp,&x,&y);
        PUTPIXS(x,y,0);
      }
      if (rp>0.0) theta+=M_PI/(30.0*rp/prada); else break;
    }
    rp+=M_PIO2/Y_RES; //
  }
  //saucer's forward ray
  rp=0.0; //
  while (rp<prada*.82)
  {
    erase=0;
    // rc = rcb +- sqrt( rcb*rcb - rcc )
    rcb = v * cos(anglook-angtra) ;
    rcd = rcb*rcb - rcc ;
    if(rcd < 0.0)
      erase=1;
    else
    {
      rcd = sqrt(rcd) ;
      rc  = rcb + rcd ;
      rc2 = rcb - rcd ;
      if ( rc*rc2<0.0 )
      { //it's always this case, rc*rc2<0.0, except for extremely high speed
          if ( rp+radialfudge >= (max(rc,rc2)) ) erase=1;
      }
      else
      if ( rc>0.0 && rc2>0.0 )
      {
        if ( rp-radialfudge <= (min(rc,rc2)) ) erase=1;
      }
      else
      if ( rc<0.0 && rc2<0.0 )
      {
        erase=1;
      }
      else
      if ( rc*rc2==0.0 )
      {
        if (rc+rc2<=0.0) erase=1;
      }
    }
    if (erase)
    {
      surfacerun(lon, lat, anglook, rp, &lonp, &latp);
      projec(lonp,latp,&x,&y);
      PUTPIXS(x,y,0);
    }
    rp+=M_PIO2/Y_RES;
  }
}

vector flatfrictionint(float fr, float vt, float w, float r1, float m1)
{ // returns ffx,ffy,tf -- force of friction x and y components (x parallel to vt) and torque of friction
 //  for flat (filled) circle of uniform density travelling (vt) and spinning (w) on a 2D surface with coefficient of friction times gravity fr
 //  numerical integration;  ffy should come out zero
 // doing rectangle rule integration for now; might change to trap or rk4 later
  int dds=12; // area integral 1D steps
  float x,y;
  float dx,dy;
  float ugp;
  int i,k,dxs,dys;
  float ffx=0.0;
  float ffy=0.0;
  float tf=0.0;
  vector frictintret;
  float flt1,flt2,flt3,flt4,flt5,flt6;

  ugp = m1 / (M_PI*r1*r1) * fr; // ugp = density * fr = u * g * p
  y = -r1;
  dys=dds; dy=2.0*r1/dys;
  for (k=0;k<dys;k++)
  {
    float dffx=0.0, dffy=0.0, dtf=0.0;
    flt1 = sqrt( r1*r1 - y*y );
    dxs=max(ceil( 2.0*flt1 /dy),1); dx=2.0*flt1/dxs;
    x = -flt1;
    flt1 = vt - w*y;
    flt3 = flt1*flt1;
    flt5 = y*flt1;
    flt6 = -ugp*dx;
    for (i=0;i<dxs;i++)
    {
      flt4 = w*x;
      flt2 = flt3 + flt4*flt4;
      if(flt2!=0.0) 
      {
        flt2 = flt6/sqrt( flt2 );
        dffx += flt1*flt2;
        //dffy += flt4*flt2;
        dtf  += ( flt4*x - flt5 ) * flt2 ;
      }
      x += dx;
    }
    ffx += dffx*dy;
    //ffy += dffy*dy;
    tf  += dtf*dy;
    y += dy;
  }
  frictintret.x = ffy ;
  frictintret.y = ffx ; // switching .x and .y returns to match up with spherefrictionint (.y is torque parallel to vt in return)
  frictintret.z = tf ;
  return frictintret;
}

vector spherefrictionint(float fr, float vt, float w, float r1, float m1)
{ // returns tf(x,y,z) -- torque of friction (x parallel to vt)
 //  for spherical shell cap of uniform density travelling (vt) and spinning (w) on a unit sphere surface with coefficient of friction times gravity fr
 //  numerical integration;
 // doing rectangle rule integration for now; might change to trap or rk4 later
  int dds=12; // area integral 1D steps
  float ddth=M_PIT2/dds;
  float ph,th;
  float dph,dth;
  float ugp;
  int i,k,dphs,dths;
  vector tf = { 0.0, 0.0, 0.0 };
  vector r,wvec,wvecxr;
  float flt1,flt2,flt3;
  ugp = m1 / (M_PIT2*(1-cos(r1))) * fr; // ugp = density * fr = u * g * p
  wvec.x=0.0; wvec.y=vt; wvec.z=w;
  ph = 0.0;
  dphs=dds; dph= r1/dphs;
  for (k=0;k<dphs;k++)
  {
    vector dtf = { 0.0, 0.0, 0.0 };
    flt2 = sin(ph);
    r.z = cos(ph);
    th = 0.0;
    dths=ceil( M_PIT2 /ddth); dth=M_PIT2/dths;
    for (i=0;i<dths;i++)
    {
      r.x = flt2 * cos(th);
      r.y = flt2 * sin(th);
      wvecxr = unitvect( vectcross( wvec, r ) );
      dtf = vectsum( dtf, vectcross( r, wvecxr ) );  // possibly could simplify this with r x (w x r) = r^2 w (how true is that? not very, eh? but...)
      th += dth;
    }
    flt1 = -ugp*dth*sin(ph)*dph;
    dtf = vecttimes( flt1 , dtf );
    tf = vectsum(tf,dtf);
    ph += dph;
  }
  return tf;
}