// Spherical Spacewar -- Spacewar on the surface of a sphere
//   Projected Mercator, Winkel Tripel, sinusoidal,
//   two hemispheres flat, or the full sphere flat and two-valued,
//   or various other projections.
//   Momentum conserved in shooting and collisions.
//   Player creates mass of shots at the time of shooting, out of thin air.
//   Shot hits increase the mass and radius of the hit player.
//   Player mass is directly proportional to player radius.
//   Player shape is a spherical shell cap (the zone of a spherical cap), 
//   or a filled circle in rectilinear mode.
//   Program has rectilinear mode (not on a sphere).
//   Rectilinear mode is wrapping flat space or walled flat space.
//   Program has option for a physics read-out of energy and momentum 
//   at bottom of screen.
// Allen Jordan, 6-15-06 to 7-31-06
#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include <mem.h>
#include <stdlib.h>
#include <math.h>
#include <values.h>
#include <time.h>
#include <process.h>
#include "sphrspwa.h"
#include "sphrspw2.h"

// Alex Russell keyboard handler Copyright 1995, May 14 by Alex Russell, ALL rights reserved
#define BYTE unsigned char
#define NUM_SCAN_QUE 256     // this MUST be 256, using BYTE roll-over for q code
static void interrupt (far *oldkb)(void);   /* BIOS keyboard handler */
BYTE gb_scan; // Q code
BYTE gb_scan_q[NUM_SCAN_QUE];
BYTE gb_scan_head;
BYTE gb_scan_tail; // /Alex Russell

unsigned char far *SCREEN=MK_FP(VIDEO,0);
unsigned char far *canvas;

void sinproj(float , float , int *, int *);  //sinproj prototype
projecpointer projec = sinproj;

projecpointer mixprojold, mixprojnew;
float mixproja; // and (1-mixproja), coefficients for linear combination of projections

void (*drawobj)(float, float, float, float, int, int);

void (*p1p2bounces)(float);

void (*playerwallbounces)(float , float , float *, float *, float *);

float (*metric)(float , float , float, float) ;

void (*surfacerun)(float , float , float , float , float *, float *) ;

void (*intermediatepoint)(float , float , float , float , float , float *, float *) ;

float (*angtofrom)(float , float , float , float ) ;

void (*moveplayer)(float *, float *, float , float *, float *, float , float) ;

float (*playermoment)(float) ;

void (*absorbknockturn)(float *, float , float , float , float , float , float) ;

vector (*frictionintegral)(float, float , float , float , float ) ;

// player control increments
float vinc=0.001; //speed increment
float ainc=0.142799666072263; // turning angle increment 2pi/44; 

// player 1
int   p1exists;
float p1lat, p1long;  // int p1[2]; // lat,long ; screen x,y
float p1a; // angle that player is facing
float p1at;  // angle of travel, angles are north of east;
float p1v; // speed
float p1av; // angular velocity
float p1rad; // player 1 saucer radius
int p1colc, p1colf; //color of center, color of forward angle
unsigned int p1shotsfired=0;

// player 2
int   p2exists;
float p2lat, p2long; //int p2[2]; // lat,long ; screen x,y
float p2a; // angle that player is facing
float p2at;  // angle of travel, angles are north of east;
float p2v; // speed
float p2av; // angular velocity
float p2rad; // player 2 saucer radius
int p2colc, p2colf; //color of center, color of forward angle
unsigned int p2shotsfired=0;

struct player
{
  int exists;
  float lon;
  float lat;
  float a; // angle that player is facing
  float bear; // angle of travel
  float v; //, vlon, vlat; //speed   
  float av; // angular velocity  //
  float rad; // player saucer radius
  int colc, colf; //color of center, color of forward angle
  unsigned int shotsfired;
} plrsaverestore[2]; 
// this struct would also be a somewhat cleaner way to do the players' active data; 
// a moderate amount of work to change to; worth changing to if adding a computer-
// controlled 3rd player, to work with an array of player structs

// shots
struct shot
{
  int owner; // 0:p1, 1:p2, -1:free
  unsigned int serial; // set to player's # shots fired, used to detonate the oldest
  float lon;
  float lat;
  float bear; // angle of travel
  float v; //speed
  float rad;
  int col;
  int fresh; //fresh in the barrel
  int exploding;  // boolean
  int stale; //exploding and has acted on a player
  int wipeshottrail; //flag used to erase the old pre-exploding shot streak after the shot starts exploding 
  int oldx[SHOTDRAWMEM];
  int oldy[SHOTDRAWMEM];
  int oldindex; //starting index into ring array of previous draws' xy's
} shots[MAX_SHOTS];
float muzzlespeed;
float shotexplrad;

//other globals;
int gridcol=60; //252;  //105;
int backproj=0; //flag for if the projected image is // viewed through from opposite side, like from under instead of above
int trajmath=0; //trajectory math method selector
float uf_p2p=.1; // coefficient of friction between players, for transverse impulse in collisions
int curioswitch2=0;  // programmers probe on '|'

  int keyp1tl,keyp1f,keyp1tr,keyp1l,keyp1b,keyp1r,keyp1fire,keyp1det; //booleans for player keys down
  int keyp2tl,keyp2f,keyp2tr,keyp2l,keyp2b,keyp2r,keyp2fire,keyp2det;
  int keyenvintcom; // ascii character of latest environment interface command key

int p1turnl=0, p2turnl=0;
int p1turnr=0, p2turnr=0;

struct 
{
  int exists; // 0 doesn't.  1 or 2, spherical shell gravity interpretations 1 or 2
  float lon;
  float lat;
  float v; // speed   
  float at; // angle of travel
  float am; // spin angular momentum  //
  float m; // mass  , in pixel units
  float rad; // event horizon radius, in angle units; formulated based on mass, rad = 2Gm/c/c
  float rad2; // player esacape horizon radius, in angle units, 2Gm/vinc/vinc
  int col; // outside event horizon lightshow color 
  float pglon[BLACKABPIXGRABS];
  float pglat[BLACKABPIXGRABS];
  float pgv[BLACKABPIXGRABS];
  float pgat[BLACKABPIXGRABS];
  int   pgc[BLACKABPIXGRABS]; //color
} blackab; 

void vga_mode(int mode)
{
  union REGS regs;

  regs.x.ax = mode;
  int86(0x10, &regs, &regs);
}

void clear_screen(void)
{
  _fmemset(canvas, 0, (unsigned)64000);
}

void vgapalset(void)
{
  int i,j,k;
  int vgapal[256][3]; //custom vga palette

  {
  k=0;
  for (j=0;j<3;j++) for (i=0;i<(16);i++){
    vgapal[i+16*j][j]= (4*i<64) ? (4*i) : 63;
    vgapal[i+16*j][(j+1)%3]=0;
    vgapal[i+16*j][(j+2)%3]=0;}
   k+=16*3; //48
  for (j=0;j<3;j++) for (i=0;i<(16);i++){
    vgapal[k+i+16*j][j]= (4*i<64) ? (4*i) : 63;
    vgapal[k+i+16*j][(j+1)%3]= (4*i<64) ? (4*i) : 63;
    vgapal[k+i+16*j][(j+2)%3]=0;}
   k+=16*3; //96
  for (i=0;i<16;i++) for (j=0;j<3;j++)
    vgapal[k+i][j]= (4*i<64) ? (4*i) : 63;
   k+=16;   //112
  for (j=0;j<3;j++) for (i=0;i<(16);i++){
    vgapal[k+i+16*j][j]= (4*i<64) ? (4*i) : 63;
    vgapal[k+i+16*j][(j+1)%3]= 2*i;
    vgapal[k+i+16*j][(j+2)%3]= 2*i;}
   k+=16*3; //164
  for (j=0;j<3;j++) for (i=0;i<(16);i++){
    vgapal[k+i+16*j][j]= (4*i<64) ? (4*i) : 63;
    vgapal[k+i+16*j][(j+1)%3]= (4*i<64) ? (4*i) : 63;
    vgapal[k+i+16*j][(j+2)%3]= 2*i;}
   k+=16*3; //208
  for (j=0;j<3;j++) for (i=0;i<(16);i++){
    vgapal[k+i+16*j][j]= (4*i<64) ? (4*i) : 63;
    vgapal[k+i+16*j][(j+1)%3]= 1.3*i;
    vgapal[k+i+16*j][(j+2)%3]= 1.3*i;}
   k+=16*3; //256
  for(j=0;j<3;j++) {vgapal[0][j]=0;}
  }
  //outp(0x03C6,0xff);
  outp(0x03c8, 0);
  for(i=0;i<256;i++)
  {
    for(j=0;j<3;j++)
    {
    outp(0x03c9, vgapal[i][j]);
    }
  }
}

// Alex Russell keyboard handler Copyright 1995, May 14 by Alex Russell, ALL rights reserved
/* ---------------------- get_scan() --------------------- April 17,1993 */
void interrupt get_scan(void)
{
   /* read the raw scan code from the keyboard */
   asm   cli
   asm   {
         in    al, 060h       /* read scan code */
         mov   gb_scan, al
         in    al, 061h       /* read keyboard status */
         mov   bl, al
         or    al, 080h
         out   061h, al       /* set bit 7 and write */
         mov   al, bl
         out   061h, al       /* write agian, bit 7 clear */
         mov   al, 020h       /* reset PIC */
         out   020h, al
         /* end of re-set code */
         sti
         }
   *(gb_scan_q+gb_scan_tail)=gb_scan;
   ++gb_scan_tail;
}
/* ---------------------- init_keyboard() ---------------- April 17,1993 */
void init_keyboard(void)
{
   BYTE far *bios_key_state;
   /* save old BIOS key board handler */
   oldkb=getvect(9);
   // turn off num-lock via BIOS
   //bios_key_state=MK_FP(0x040, 0x017);
   //*bios_key_state&=(~(32 | 64));     // toggle off caps lock and
                                      // num lock bits in the BIOS varible
   //oldkb();      // call BIOS keyhandler to change keyboard lights
   gb_scan_head=0;
   gb_scan_tail=0;
   gb_scan=0;
   /* setup our own handler */
   setvect(9, get_scan);
}
/* restore the bios keyboard handler */
/* ---------------------- deinit_keyboard() -------------- April 17,1993 */
void deinit_keyboard(void)
{
   setvect(9, oldkb);                             // /Alex Russell
}

unsigned char ctrlaltorshift(unsigned char keysc)
{
  return ( keysc==0x1d || keysc==0x38 || keysc==0x2a || keysc==0x36 );
}

int keyboardproc()
// reads and flushes keyboard buffer for UI commands
//  sets player control flags and interface key message
// returns 1 or 2 if any keys down were in the buffer, except ctl alt or shift;
//   2 if there was a player control key down
// returns 0 if no keys, except ctl alt or shift, or only key ups
{
  static unsigned char lshifton=0, rshifton=0;
  unsigned char shifton;
  unsigned char ch=0;
  unsigned char keysc;
  unsigned char anykeys=0;

  while ( gb_scan_head != gb_scan_tail )
  {
    keysc=*(gb_scan_q+gb_scan_head);
    ++gb_scan_head;

    if( !anykeys && keysc<0x80 && keysc!=0x1d && keysc!=0x38 && keysc!=0x2a && keysc!=0x36 ) anykeys=1;
    if( lshifton || rshifton ) shifton=1; else shifton=0;

    switch( keysc )
    {
      case 0x2a :   lshifton=1; break;
      case 0xaa :   lshifton=0; break;
      case 0x36 :   rshifton=1; break;
      case 0xb6 :   rshifton=0; break;

      case 0x01 :   ch=0x1b ; break;
      case 0x06 :   if(shifton) ch='%' ; else ch='5' ; break;
      case 0x07 :   if(shifton) ch='^' ; else ch='6' ; break;  
      case 0x19 :   if(shifton) ch='P' ; else ch='p' ; break;
      case 0x1a :   if(shifton) ch='{' ; else ch='[' ; break;
      case 0x1b :   if(shifton) ch='}' ; else ch=']' ; break;
      case 0x14 :   if(shifton) ch='T' ; else ch='t' ; break;
      case 0x25 :   if(shifton) ch='K' ; else ch='k' ; break; 
      case 0x16 :   if(shifton) ch='U' ; else ch='u' ; break;
      case 0x27 :   if(shifton) ch=':' ; else ch=';' ; break; 
      case 0x28 :   if(shifton) ch='"' ; else ch='\'' ; break; 
      case 0x39 :   ch=' ' ; break;
      case 0x35 :   if(shifton) ch='?' ; else ch='/' ; break;
      case 0x13 :   if(shifton) ch='R' ; else ch='r' ; break; 
      case 0x2b :   if(shifton) ch='|' ; else ch='\\' ; break;
      case 0x26 :   if(shifton) ch='L' ; else ch='l' ; break; 
      case 0x18 :   if(shifton) ch='O' ; else ch='o' ; break; 
      case 0x17 :   if(shifton) ch='I' ; else ch='i' ; break; // i free
      case 0x15 :   if(shifton) ch='Y' ; else ch='y' ; break; // Y y free
      case 0x22 :   if(shifton) ch='G' ; else ch='g' ; break; 
      case 0x2f :   if(shifton) ch='V' ; else ch='v' ; break; 
      case 0x30 :   if(shifton) ch='B' ; else ch='b' ; break;
      case 0x23 :   if(shifton) ch='H' ; else ch='h' ; break; 
      case 0x0c :   if(shifton) ch='_' ; else ch='-' ; break; // _ free
      case 0x29 :   if(shifton) ch='~' ; else ch='`' ; break; 
      case 0x0d :   if(shifton) ch='+' ; else ch='=' ; break; // + free
      case 0x0b :   if(shifton) ch=')' ; else ch='0' ; break; // ) free
      case 0x02 :   if(shifton) ch='!' ; else ch='1' ; break; // free
      case 0x03 :   if(shifton) ch='@' ; else ch='2' ; break; // free
      case 0x08 :   if(shifton) ch='&' ; else ch='7' ; break; 
      case 0x09 :   if(shifton) ch='*' ; else ch='8' ; break; 

      case 0x2c :   keyp2tl=1; anykeys=2; break;
      case 0x2d :   keyp2f=1; anykeys=2; break;
      case 0x2e :   keyp2tr=1; anykeys=2; break;
      case 0x1e :   keyp2l=1; anykeys=2; break;
      case 0x1f :   keyp2b=1; anykeys=2; break;
      case 0x20 :   keyp2r=1; anykeys=2; break;
      case 0x12 :   keyp2fire=1; anykeys=2; break;
      case 0x11 :   keyp2det=1; anykeys=2; break;

      case 0xac :   keyp2tl=0; break;
      case 0xad :   keyp2f=0; break;
      case 0xae :   keyp2tr=0; break;
      case 0x9e :   keyp2l=0; break;
      case 0x9f :   keyp2b=0; break;
      case 0xa0 :   keyp2r=0; break;
      case 0x92 :   keyp2fire=0; break;
      case 0x91 :   keyp2det=0; break;

      case 0x4f :   keyp1tl=1; anykeys=2; break;
      case 0x50 :   keyp1f=1; anykeys=2; break;
      case 0x51 :   keyp1tr=1; anykeys=2; break;
      case 0x4b :   keyp1l=1; anykeys=2; break;
      case 0x4c :   keyp1b=1; anykeys=2; break;
      case 0x4d :   keyp1r=1; anykeys=2; break;
      case 0x49 :   keyp1fire=1; anykeys=2; break;
      case 0x48 :   keyp1det=1; anykeys=2; break;

      case 0xcf :   keyp1tl=0; break;
      case 0xd0 :   keyp1f=0; break;
      case 0xd1 :   keyp1tr=0; break;
      case 0xcb :   keyp1l=0; break;
      case 0xcc :   keyp1b=0; break;
      case 0xcd :   keyp1r=0; break;
      case 0xc9 :   keyp1fire=0; break;
      case 0xc8 :   keyp1det=0; break;
    }
    if (ch==0x1b) keyenvintcom = ch;
  }
  if ( keyenvintcom!=0x1b )  keyenvintcom = ch ;
  return anykeys;
}

void oldpauser(void)
{
  while (!kbhit()) ;
  getch();
}

void pauser(int *pause)
{
  int i;
  while ( !(i=keyboardproc()) );
  if (keyenvintcom==' ' || i==2) 
  {
    *pause=0;
    keyenvintcom=0;
  }
}

void textpauser(int *pause, clock_t *keytick)
{
  int i;
  gotoxy(1,24);
  if (*pause)
  {
    printf(" (space)");
    gotoxy(79,24);
    while ( !(i=keyboardproc()) );
    if (keyenvintcom==' ' || i==2) 
    {
      keyenvintcom=0;
      *pause=0;
      *keytick = clock();
    }
  }
}

void errnote(void)
{
  int pausdum;
  printf("Problem.  Press any key.\n");
  pauser(&pausdum);
}

void curiopauser(void)
{
  int ch;
  while ( !keyboardproc() );
  if(keyenvintcom==' ')keyenvintcom=0; else curioswitch2=0;
}

int yres(int y)
{
  if (y<0) return 0;
  if (y>=200) return 199;
  return y;
}

int imod( int a, int b )
//integer mod, no negative returns
{
  a = a % b ; 
  if (a<0) a += b;
  return a;
}

float twopimod( float angle )
{
  float answer;
  answer = fmod( angle , M_PIT2 );
  if ( answer<0.0 ) answer += M_PIT2;
  return answer;
}

float pimod( float angle )
{
  float answer;
  answer = fmod( angle , M_PI );
  if ( answer<0.0 ) answer += M_PI;
  return answer;
}

float nearfmodmult( float quant, float grain )
{ //quantizes quantity quant to nearest multiple of grain 
  //this should work, but test this function
  float flt;
  flt = fmod(quant,grain) ;
  if(flt==0.0) return quant;
  else
  {
    if(quant>0.0)
    {
      if ( 2.0*flt>=grain )           // fabs(grain) ) ; assume positive grain
        return quant+grain-flt;
      else
        return quant-flt;
    }
    if(quant<0.0)
    {
      if ( -2.0*flt>=grain )           // fabs(grain) ) ; assume positive grain
        return quant-grain-flt;
      else
        return quant-flt;
    }
  }
}

void sinproj(float lon, float lat, int *x, int *y)
{
// sinusoidal projection
//  takes longitude and latitude; converts to screen x and y
  *y = Y_OFFS-0.45 - lat / M_PI * Y_RESM*0.9;    //( 0.5 - lat / M_PI ) * Y_RESM;
  *x = ( lon * cos(lat) / M_PIT2 + 0.5 ) * X_RES;
  if (*y<0) *y=0; if (*y>=Y_RES) *y=Y_RES-1;
  if (*x<0) *x=0; if (*x>=X_RES) *x=X_RES-1;
  backproj=0;
}

void winktrproj(float lon, float lat, int *x, int *y)
{
// Winkel Tripel projection
//  takes longitude and latitude; converts to screen x and y
  float a,w,lato;
  float flt1,flt2; //temp floats
  lato = 0.880689235420356667; // acos(2.0/M_PI);
  a = acos(cos(lat)*cos(lon/2.0));
  flt1=sin(a); if (flt1==0.0) w = 0; else w = 1.0/flt1;
  flt1 = (lat + w*a*sin(lat))/2;
  *y =  Y_OFFS -0.45 - flt1 / M_PI * Y_RESM*.9 ;// ( M_PI/2.0 - flt1 ) / M_PI * Y_RESM ;//*.95+5;
  flt1 = (lon*cos(lato)+2.0*w*a*cos(lat)*sin(lon/2.0))/2.0;
  *x = ( flt1*1.218 + M_PI ) / M_PIT2 * X_RES;
  if (*y<0) *y=0; if (*y>=Y_RES) *y=Y_RES-1;
  if (*x<0) *x=0; if (*x>=X_RES) *x=X_RES-1;
  backproj=0;
}

void merctrproj(float lon, float lat, int *x, int *y)
{
// Mercator projection
#define MERCSCALEOPI 0.265258238486492226 // scale 1/1.2 over pi
#define MERCLIM 1.55  //improve this ..
  if (lat>MERCLIM) *y=0;
  else
  if (lat<-MERCLIM) *y=Y_RESM;
  else
  {
    lat = log(tan(M_PI_4+lat/2.0));
    *y = ( 0.5 - lat * MERCSCALEOPI ) * Y_RESM ;  
  }
  *x = ( lon + M_PI ) / M_PIT2 * X_RES;
  if (*y<0) *y=0; if (*y>=Y_RES) *y=Y_RESM;
  if (*x<0) *x=0; if (*x>=X_RES) *x=X_RESM;
  backproj=0;
}

void rectproj(float lon, float lat, int *x, int *y)
{
// rectangular projection -- x and y directly proportional to lon and lat
//  takes longitude and latitude; converts to screen x and y
  *y = Y_OFFS-0.45 - lat / M_PI * (Y_RESM)*.9;   // ( 0.5 - lat / M_PI ) * (Y_RESM);
  *x = ( lon / M_PIT2 + 0.5 ) * (X_RES);
  if (*y<0) *y=0; if (*y>=Y_RES) *y=Y_RESM;
  if (*x<0) *x=0; if (*x>=X_RES) *x=X_RESM;
  backproj=0;
}

void twohemiproj(float lon, float lat, int *x, int *y)
{
// Two Hemispheres projection
//   two hemisphere split at lon=0,pi, rotated to face screen
//  takes longitude and latitude; converts to screen x and y
  float c3x,c3y,c3z;
  float flt1=.95;
  float coslat;
  coslat=cos(lat);
  c3x = cos(lon)*coslat; //screen x '
  c3y = sin(lon)*coslat; //transverse to screen
  c3z = sin(lat); //screen y '

  //*y = ( 1.0 - c3z ) * Y_RESM/3.0 + Y_RESM/6.0;
  *y = ( 0.5 - c3z*0.33333333333 ) * Y_RESM;

  *x = ( c3x*flt1 + 1.0 ) * X_RES * 0.25 ;
  if (c3y>0.0)
    *x = X_RES - *x ;

  if (*y<0) *y=0; if (*y>=Y_RES) *y=Y_RES-1;
  if (*x<0) *x=0; if (*x>=X_RES) *x=X_RES-1;
  backproj=0;
}

void twohemiproj2(float lon, float lat, int *x, int *y)
{
// Two Hemispheres projection
//   two hemisphere split at lon=0,pi, rotated to face / face-away-from screen
//    (hemisphere on the right seen through from under it)
//  takes longitude and latitude; converts to screen x and y
  float c3x,c3y,c3z;
  float flt1=.95;
  float coslat;
  coslat=cos(lat);
  c3x = cos(lon)*coslat; //screen x '
  c3y = sin(lon)*coslat; //transverse to screen
  c3z = sin(lat); //screen y '

  //*y = ( 1.0 - c3z ) * Y_RES/3.0 + Y_RES/6.0;
  *y = ( 0.5 - c3z/3.0 ) * Y_RESM;

  *x = ( c3x*flt1 + 1.0 ) * X_RES*0.25;
  if (c3y<=0.0)
    backproj=0;
  else
  { *x += X_OFFS+1;    backproj=1; }

  if (*y<0) *y=0; if (*y>=Y_RES) *y=Y_RES-1;
  if (*x<0) *x=0; if (*x>=X_RES) *x=X_RES-1;
}

void sphereproj(float lon, float lat, int *x, int *y)
{
// Flat projection of Sphere to plane
//   back hemisphere seen through from under it
//  takes longitude and latitude; converts to screen x and y
  float c3x,c3y,c3z;
  float flt1=.9;
  c3x = cos(lon)*cos(lat); //transverse to screen
  c3y = sin(lon)*cos(lat); //screen x '
  c3z = sin(lat); //screen y '

  *y = ( 1.0 - c3z ) * Y_RESM*0.5;

  *x = ( c3y*flt1 ) * X_RES*0.5*flt1 + X_OFFS;
  if (c3x>=0.0)
    backproj=0;
  else
  {
    backproj=1;
    *x=(*x)-1;
  }

  if (*y<0) *y=0; if (*y>=Y_RES) *y=Y_RES-1;
  if (*x<0) *x=0; if (*x>=X_RES) *x=X_RES-1;
}

void azimuequiproj(float lon, float lat, int *x, int *y)
{
// azimuthal equidistant projection
  float c,kp;
  float coslat,sinc;
//  float fx,fy;

  coslat = cos(lat);
  c = acos( coslat * cos(lon) ) ;
  sinc = sin(c) ;
  if (sinc!=0.0) kp =  c / sinc ; else if(c==0.0) kp = 1; else kp = 1000000;
  coslat = kp * coslat * sin(lon ) ;
  sinc = kp * sin(lat) ;
  *x = ( coslat / 6 + 0.5 ) * X_RES ;
  *y = ( 0.5 - sinc/5 ) * Y_RESM ;
  backproj=0;
  if (*y<0) *y=0; if (*y>=Y_RES) *y=Y_RESM;
  if (*x<0) *x=0; if (*x>=X_RES) *x=X_RESM;
}

void albersconicproj(float lon, float lat, int *x, int *y)
{
// Albers Equal-Area Conic Projection
#define ALBERSN   0.433012701892219323381  // n = sqrt(3)/4.0 ;
#define ALBERSNT2 0.866025403784438646764  // 2*n
#define ALBERSN1O 2.30940107675850305804  // p0 = 1.0 / n ;
  float th,p;
  float fx,fy;
  th = ALBERSN*lon ; 
  p = sqrt( 1.0 - ALBERSNT2*sin(lat) ) ; 
  fx = p * sin(th) ;
  fy = ALBERSN1O - p * cos(th) ;
  *x = ( fx * .36 + 0.5 ) * X_RES ;
  *y = ( 1.4 - fy * .55 ) * Y_RESM ;
  backproj=0;
  if (*y<0) *y=0; if (*y>=Y_RES) *y=Y_RESM;
  if (*x<0) *x=0; if (*x>=X_RES) *x=X_RESM;
}

void perspsphereproj(float lon, float lat, int *x, int *y)
{
// perspective projection of Sphere to plane
//   back hemisphere seen through from under it
//  takes longitude and latitude; converts to screen x and y
  float c3x,c3y,c3z;
  float flt1=.81;
  c3x = cos(lon)*cos(lat); //transverse to screen
  c3y = sin(lon)*cos(lat); //screen x '
  c3z = sin(lat); //screen y '

  *y = Y_OFFS-0.45 - c3z/(8.01-c3x) * Y_RESM*0.5*7.01;

  *x = ( c3y / (8.01-c3x) ) * X_RES*0.5*flt1*7.01 + X_OFFS;
  if (c3x>=0.0)
    backproj=0;
  else
  {
    backproj=1;
  }
  if (*y<0) *y=0; if (*y>=Y_RES) *y=Y_RES-1;
  if (*x<0) *x=0; if (*x>=X_RES) *x=X_RES-1;
}

void mixproj(float lon, float lat, int *x, int *y)
// mixed projection, linear combination of two other projections
// for animated transition from one to the other
// accepts additional parameters, by global variables, 
//  mixproja and (1-mixproja), coefficients for the combination
// and two global projection function pointers, mixprojold and mixprojnew
{
  int x2,y2,backproj2;
  mixprojold(lon,lat, x, y);
  backproj2=backproj;
  mixprojnew(lon,lat,&x2,&y2);
  *x = *x*mixproja + x2*(1.0-mixproja) ;
  *y = *y*mixproja + y2*(1.0-mixproja) ;
//  backproj=0;  // as to telling when a point on a mixed projection is backproj,
               // could test for lon+ causing x-; would take twice as long
// but excepting a real test, and neglecting those cases where a combination of two
// non-backprojected points might be back-projected, this:
  if(mixproja>.5) backproj=backproj2;
}

void squareproj(float lx, float ly, int *x, int *y)
{
// projection function for when in rectilinear mode -- not sphere surface
// lon-x, lat-y go to 2pi by pi letterboxed rectangle of screen x,y
// ? letterbox y range to 160 or 133 ?
  float ylet=133;
  *y = ( 0.5 - ly / M_PI ) * ylet + 30;
  *x = ( lx / M_PIT2 + 0.5 ) * (X_RES);
  if (*y<0) *y=0; if (*y>=Y_RES) *y=Y_RESM;
  if (*x<0) *x=0; if (*x>=X_RES) *x=X_RESM;
  backproj=0;
}

float spheremetric(float lon1, float lat1, float lon2, float lat2)
// distance on a sphere between two points in lon,lat
//  a sphere of radius 1.0
{
  float flt;
  flt = cos(lat1)*cos(lat2)*cos(lon1-lon2) + sin(lat1)*sin(lat2);
//  if (fabs(flt)<=1.0)
    return ( acos( flt ) );
//  else return M_PI;
}

/*
float spheremetric2(float lon1, float lat1, float lon2, float lat2)
// distance on a sphere between two points in lon,lat
//  a sphere of radius 1.0
// equivalent to spheremetric but said to be less subject to rounding errors over short distances
// at http://williams.best.vwh.net/avform.htm#Dist
{
  float flt1,flt2;
  flt1 = sin((lat1-lat2)/2.0);
  flt2 = sin((lon1-lon2)/2);
  return ( 2.0 * asin( sqrt( flt1*flt1 + cos(lat1)*cos(lat2)*flt2*flt2)) );
//d=2*asin( sqrt( (sin((lat1-lat2)/2))^2 + cos(lat1)*cos(lat2)*(sin((lon1-lon2)/2))^2 ) )
}
*/

void sphererun(float lon1, float lat1, float angtra, float d, float *lon2, float *lat2)
{
// returns lon,lat of point distance d out on a great circle path starting with bearing angtra from lon1,lat1
// from http://williams.best.vwh.net/avform.htm#LL
  float flt1,flt2,flt3,flt4;
  flt3 = cos(lat1)*sin(d);
  flt4 = sin(lat1);
  flt2 = cos(d);
  flt1 = cos(angtra)*flt3;
  flt3 = flt4*flt2 + flt3*sin(angtra);
  *lat2 = asin(flt3);
  flt2 = flt2-flt4*flt3;
  if (flt1!=0.0 || flt2!=0.0) flt1=atan2(flt1,flt2); else flt1=0.0;
  *lon2= twopimod( lon1 + flt1 + M_PI) - M_PI ;
}
//     lat =asin(sin(lat1)*cos(d)+cos(lat1)*sin(d)*cos(tc))
//     dlon=atan2(sin(tc)*sin(d)*cos(lat1),cos(d)-sin(lat1)*sin(lat))
//     lon=fmod( lon1-dlon +pi,2*pi )-pi

float squaremetric(float x1, float y1, float x2, float y2)
// distance on a wrapping rectangle between two points
{
  float dx1,dx2,dy1,dy2,dret,flt;
  dx1=x2-x1;
  dy1=y2-y1;
  if ( dy1==0.0 && dx1==0.0 ) return 0.0;
  if(dx1>0.0) 
    dx2=dx1-M_PIT2;
  else
    dx2=dx1+M_PIT2;
  if(dy1>0.0) 
    dy2=dy1-M_PI;
  else
    dy2=dy1+M_PI;
  dret = dx1*dx1 + dy1*dy1;
  flt = dx2*dx2 + dy1*dy1; if(flt<dret)dret=flt;
  flt = dx2*dx2 + dy2*dy2; if(flt<dret)dret=flt;
  flt = dx1*dx1 + dy2*dy2; if(flt<dret)dret=flt;
  return sqrt(dret);
}

void squarerun(float x1, float y1, float angtra, float d, float *x2, float *y2)
{
// returns x,y of point distance d out on wrapping rectangle path starting with bearing angtra from x1,y1
// rectilinear version of sphererun(...)
  *x2 = x1 + d * cos(angtra) ;
  *y2 = y1 + d * sin(angtra) ;
  *x2 = twopimod( *x2 + M_PI) - M_PI ;
  *y2 = pimod( *y2 + M_PIO2) - M_PIO2 ;
}

float walledsquaremetric(float x1, float y1, float x2, float y2)
// distance on a non-wrapping rectangle between two points
{
  float dx1,dy1,dret;
  dx1=x2-x1;
  dy1=y2-y1;
  dret = dx1*dx1 + dy1*dy1;
  return sqrt(dret);
}

void walledsquarerun(float x1, float y1, float angtra, float d, float *x2, float *y2)
{
// returns x,y of point distance d out, or to a wall, in a walled rectangle space, path starting with 
// bearing angtra from x1,y1
// walled version of squarerun(...)
  *x2 = x1 + d * cos(angtra) ;
  *y2 = y1 + d * sin(angtra) ;
  if(*x2>M_PI) *x2= 3.12588;   // M_PI; // or --  3.12588 
  if(*x2<-M_PI) *x2= -3.12588;   // -M_PI;
  if(*y2>M_PIO2) *y2= 1.56215; // M_PIO2;  // or --   1.56215
  if(*y2<-M_PIO2) *y2= -1.56215; // -M_PIO2;
}

vector cartr(float lon1, float lat1)
{
  vector r1;
  r1.x = cos(lon1)*cos(lat1);
  r1.y = sin(lon1)*cos(lat1);
  r1.z = sin(lat1);
  return r1;
}

vector cartv(float lon, float lat, float spd, float angtra)
{
  vector v;
  v.x = spd*(-cos(angtra)*sin(lon)-sin(angtra)*cos(lon)*sin(lat));
  v.y = spd*(cos(angtra)*cos(lon)-sin(angtra)*sin(lon)*sin(lat));
  v.z = spd*(sin(angtra)*cos(lat));
  return v;
}

vector cartd(float lon1, float lat1, float lon2, float lat2)
{
  return vectdiff(cartr(lon1,lat1),cartr(lon2,lat2));
}

vector rcrossv(float lon, float lat, float spd, float angtra)
{
  vector r,v;
  r = cartr(lon,lat);
  v = cartv(lon,lat,spd,angtra);
  v = vectcross(r,v);
  return v;
}

vector wcrossr(vector worbit, float lon, float lat)
{
  vector r;
  r = cartr(lon,lat);
  r = vectcross(worbit,r);
  return r;
}

int notonborder(float lon, float lat)
{
#define PILONBORDER  3.12 //5 //8 // 3.1258846903218442722703301663631 //3.14159
#define PIO2LATBORDER  1.568 //1.57
// returns 0 if a point is on a grid border line, 1 otherwise
// only dealing with lon=+-pi grid borders for now
  if ( projec==azimuequiproj ) return 1;
  if ( projec==sphereproj || projec==perspsphereproj ) return 1;
  if ( lon >= PILONBORDER || lon <= -PILONBORDER ) return 0;
  else return 1;
}

int notonorbyborderxy(int x, int y)
{
// returns 0 if an xy point is on or adjacent to a grid border line, 1 otherwise
// only works for horizontal/vertical lat/lon lines for now
//  (only wanted for rectilinear mode anyway )
  int xb,yb;
  if ( projec!=squareproj ) return 1;
  projec(M_PI,0,&xb,&yb); //could precalculate all four of these
  if (x+1>=xb) return 0;
  projec(-M_PI,0,&xb,&yb);
  if (x-1<=xb) return 0;
  projec(0,M_PIO2,&xb,&yb);
  if (y-1<=yb) return 0;
  projec(0,-M_PIO2,&xb,&yb);
  if (y+1>=yb) return 0;
  return 1;
}

void bresenlinengc(int x0, int y0, int x1, int y1, int c)
{
  int steep;
  int deltax;
  int deltay;
  int error;
  int ystep;
  int x,y,t;
  steep = ( abs(y1 - y0) > abs(x1 - x0) ) ;
  if (steep) {
    t=x0; x0=y0; y0=t;
    t=x1; x1=y1; y1=t;  }
  if (x0 > x1) {
    t=x0; x0=x1; x1=t;
    t=y0; y0=y1; y1=t;  }
  deltax = x1 - x0;
  deltay = abs(y1 - y0);
  error = 0;
  y = y0;
  if (y0 < y1) ystep = 1; else ystep = -1;
  for(x=x0;x<=x1;x++)
  { 
    if (steep) { if(gridcol!=GETPIXS(y,x) ||gridcol==0) PUTPIXS(y,x,c); }
    else { if(gridcol!=GETPIXS(x,y) ||gridcol==0) PUTPIXS(x,y,c); }
    error = error + deltay;
    if ( error<<1 >= deltax )  {
      y = y + ystep;
      error = error - deltax; }
  }
}

void dpnsrerasering(float lon, float lat, float anglook, float prad, int colc, int colf)
// args: lon, lat, bearing, saucer radius, color of center, color of bearing line
// a black ring around the saucer for possible use as a trail erase method
{
  float theta, rp; //parameters swept to draw the saucer
  float lonp, latp;  //lon,lat parameters, functions of theta, rp
  int x,y,xo=-1,yo=-1,xr,yr;
  float prada;
  int col = colc, colo=colc;
  float coslatp, thinc;

  prada=prad*M_PI/Y_RES; // saucer radius in pixels converted to angle
  if(colc==0) rp=prada + M_PI/Y_RES; 
  else { rp=prada - M_PI/Y_RES; colo = colc-3; }
  thinc=min(max(0.02220660990245105689/prada,0.10471975511965977462),0.261799387799);
  theta=0.0;
  while (theta<M_PIT2)
  {
    surfacerun(lon, lat, theta, rp, &lonp, &latp);
    projec(lonp,latp,&x,&y);
    if(colc!=0) { if(backproj) { col=colc-BACKPROJDIFF; colo=col-3; } else { col=colc; colo=col-3; } }
    if (xo==-1)
    {
      xr=x; yr=y;
    }
    else
    {
      if((abs(xo-x)<60)&&(abs(yo-y)<60)) 
      {
        bresenlinengc(xo,yo,x,y,col);  
        // /*   //and outie fat
        if(theta>3*M_PIO2) { if(yo!=Y_RESM&&x!=X_RESM)bresenlinengc(xo,yo+1,x+1,y,colo); }
        else if(theta>M_PI) { if(xo!=0&&y!=Y_RESM)bresenlinengc(xo-1,yo,x,y+1,colo); }
        else if(theta>M_PIO2) { if(yo!=0&&x!=0)bresenlinengc(xo,yo-1,x-1,y,colo); }
        else { if(xo!=X_RESM&&y!=0)bresenlinengc(xo+1,yo,x,y-1,colo); }
        // */
      }
    }
    xo=x; yo=y;
    //theta+=M_PI/30.0;
    theta+=thinc;
  }
  if((abs(xr-x)<30)&&(abs(yr-y)<30)) 
    bresenlinengc(x,y,xr,yr,col);
  if(colc!=0) // dpnsrerasering can be called for a player draw not erase 
  {
    surfacerun(lon, lat, anglook, rp, &lonp, &latp);    
    projec(lonp,latp,&x,&y);
    if(notonorbyborderxy(x,y)) 
    {
      PUTPIXS(x,y,colf);
      if (x!=0 && y!=0 && x!=X_RESM && y!=Y_RESM ) 
      {
        if(anglook>3*M_PIO2)    { PUTPIXS(x,y+1,colf-2); PUTPIXS(x+1,y,colf-2); }
        else if(anglook>M_PI)   { PUTPIXS(x,y+1,colf-2); PUTPIXS(x-1,y,colf-2); }
        else if(anglook>M_PIO2) { PUTPIXS(x,y-1,colf-2); PUTPIXS(x-1,y,colf-2); }
        else                    { PUTPIXS(x,y-1,colf-2); PUTPIXS(x+1,y,colf-2); }
      }
    }    
  }
}

void drawplayerring(float lon, float lat, float anglook, float prad, int colc, int colf)
{
  dpnsrerasering(lon, lat, anglook, prad - 0.4, colc, colf);
  dpnsrerasering(lon, lat, anglook, prad - 5, 0, 0);
}

void drawplayernsr(float lon, float lat, float anglook, float prad, int colc, int colf)
// args: lon, lat, bearing, saucer radius, color of center, color of bearing line
// like drawplayern but using the surfacerun() point at bearing and distance function
{
  float theta, rp; //parameters swept to draw the saucer
  float lonp, latp;  //lon,lat parameters, functions of theta, rp
  int x,y;  // screen xy of saucer point
  float rrorr;
  float prada,prada2;
  int col = colc;
  float thetasalt;
  float coslatp;

  projec(lon,lat,&x,&y);
  if (colc!=0) if (backproj) { colc-=BACKPROJDIFF; colf-=BACKPROJDIFF; }  //

  //saucer
  rp=0.0; //
  prada=prad*M_PI/Y_RES; // saucer radius in pixels converted to angle
  prada2=prada*prada;
  while (rp<=prada)
  {
    rrorr = rp*rp/prada2;
    thetasalt = ( (int)( (lon + lat + 60*rp)*400 )%36 )*M_PI/18.0; // salt so the image doesn't show arms for low theta and not higher theta
    theta=thetasalt;
    while (theta<(M_PIT2+thetasalt))
    {
      surfacerun(lon, lat, theta, rp, &lonp, &latp);
      if(notonborder(lonp,latp))
      {
        projec(lonp,latp,&x,&y);
        if (colc!=0) if(prad>=1.0) col=colc-(rrorr)*8;
        PUTPIXS(x,y,col);
      }
      if (rp>0.0) theta+=M_PI/(30.0*rp/prada); else break;
    }
    rp+=M_PIO2/Y_RES; //
  }
  //saucer's forward ray
  if(prad>=1.0)
  {
  if (col!=0) col= colf-8;
  rp=0.0; //
  while (rp<prada*.82)
  {
    surfacerun(lon, lat, anglook, rp, &lonp, &latp);
    if(notonborder(lonp,latp))
    {
      projec(lonp,latp,&x,&y);
      if (col!=0) if (col<colf) col= colf-8+rp/prada*7;
      PUTPIXS(x,y,col);
    }
    rp+=M_PIO2/Y_RES;
  }
  }
}

void eraseshottrailall(int shot)
{
  int i,x,y; 
  for(i=0;i<SHOTDRAWMEM;i++)
  {
    if ( (x=shots[shot].oldx[i])!=-1 ) 
    {
      y = shots[shot].oldy[i];
      PUTPIXS(x,y,0);
    }
  }
}

void eraseshottraillast(int shot)
{
  int i,j,x,y; 
  j=(shots[shot].oldindex + SHOTDRAWMEM-1)%SHOTDRAWMEM;
  if ( (x=shots[shot].oldx[j])!=-1 ) 
  {
    y = shots[shot].oldy[j];
    PUTPIXS(x,y,0);
  }
  else
  {
    i = 0;
    while (i<SHOTDRAWMEM)
    {
      j=(shots[shot].oldindex+i)%SHOTDRAWMEM;
      x = shots[shot].oldx[j]; if (x==-1) break;
      y = shots[shot].oldy[j];
      i++;
    }
    j=(j+SHOTDRAWMEM-1)%SHOTDRAWMEM;
    x = shots[shot].oldx[j];
    PUTPIXS(x,y,0);
  }
}

void drawshot(int shot, int erasing)
{
  float theta, rp; //parameters swept to draw the shot ball
  float lonp, latp;  //lon,lat parameters, functions of theta, rp
  int x,y,x2,y2;  // screen xy of point
  float rrorr;
  float rada,rada2;
  int colc = shots[shot].col;
  int col,cold;
  int i,j,il;
  float shotlon,shotlat;

  shotlon = shots[shot].lon;
  shotlat = shots[shot].lat;
  projec(shotlon,shotlat,&x,&y);
  if (erasing) colc=0;
  else if (backproj) { colc-=BACKPROJDIFF; }  //
  cold=col=colc;
  if (!(shots[shot].exploding)) 
  {
    if(notonborder(shotlon,shotlat))
    {
      if (!erasing) PUTPIXS(x,y,108); else PUTPIXS(x,y,col);
      i = 0;
      if (erasing) il=SHOTDRAWMEM; else il=SHOTDRAWMEM-1;
      while (i<il)
      {
        j=(shots[shot].oldindex+i)%SHOTDRAWMEM;
        x2 = shots[shot].oldx[j]; if (x2==-1) break;
        y2 = shots[shot].oldy[j];
        //if ( x2<0 || x2>=X_RES || y2<0 || y2>=Y_RES ) errnote();  //
        if ( (x2!=x || y2!=y) ) PUTPIXS(x2,y2,col);
        if (colc%16>=3) col-=3;
        i++;
      }
      j = shots[shot].oldindex;
      if ( x!=shots[shot].oldx[j] || y!=shots[shot].oldy[j] )
      {
        i = (j + SHOTDRAWMEM - 1)%SHOTDRAWMEM;
        shots[shot].oldx[i] = x;
        shots[shot].oldy[i] = y;
        shots[shot].oldindex = i;
      }
    }
  }
  else
  {
  //ball explosion
  rp=0.0; //
  rada=shots[shot].rad*M_PI/Y_RES; // radius in pixels converted to angle
  rada2=rada*rada;
  while (rp<=rada)
  {
    rrorr = rp*rp/rada2;
    theta=0.0;
    while (theta<2*M_PI)
    {
      surfacerun(shotlon, shotlat, theta, rp, &lonp, &latp);
      projec(lonp,latp,&x,&y);
      if (colc!=0)
        { col=colc-(rrorr)*8; cold=col-1; }
      if(notonorbyborderxy(x,y)) 
      {
        PUTPIXS(x,y,col);
        if (x!=0 && y!=0 && x!=X_RESM && y!=Y_RESM ) 
        {
          PUTPIXS(x,y-1,cold);
          PUTPIXS(x,y+1,cold);
          PUTPIXS(x-1,y,cold);
          PUTPIXS(x+1,y,cold); 
        }
      }
      if (rp>0.0) theta+=M_PI/(15.0*rp/rada); else break;
    }
    rp+=M_PI/Y_RES; //
  }
  }
}

void drawplayer(float lon, float lat, float anglook, float prad, int colc, int colf)
// args: lon, lat, bearing, saucer radius, color of center, color of bearing line
{
  int i,j;
  float fi,fj, iinc,jinc;
  int x,y;
  int xb, yb; //screen boundary overflow avoidance
  float ps2;
  int col = colc; //color for gradient

  projec(lon,lat,&x,&y);
  if (backproj) { colc-=BACKPROJDIFF; colf-=BACKPROJDIFF; }

  for (i=-prad;i<=prad;i++)
    for (j=-prad;j<=prad;j++)
    {
      ps2=(i*i+j*j);
      if(col!=0)   // if not erasing
        if(prad>=1.0)
        col=colc-(ps2/prad/prad)*8;
      if (ps2<=prad*prad)
      {
        xb=x+i;
        yb=y+j;
        if ((xb>=0)&&(xb<X_RES)&&(yb>=0)&&(yb<Y_RES))
        PUTPIXS(xb,yb,col);
      }
    }
  if(prad>=1.0)
  {
  if(colc==0) col=0;  // erasing
  else col= colf-8;
  fi=0.0;  fj=0.0;
  xb=x;  yb=y;
  while (fi*fi+fj*fj<prad*prad*.68)
  {
    if ((xb>=0)&&(xb<X_RES)&&(yb>=0)&&(yb<Y_RES))
      PUTPIXS(xb,yb,col);
    if(col!=0)   // if not erasing
      if (col<colf) col++;
    if (backproj) fi-=cos(anglook); else  fi+=cos(anglook);
    fj-=sin(anglook);
    xb=x+fi;
    yb=y+fj;
  }
  }
}

void sphereintermediatepoint(float lon1, float lat1, float lon2, float lat2, float fraction, float *loni, float *lati)
{
// from http://williams.best.vwh.net/avform.htm#Intermediate
  float a,b,d,x,y,z;

  d = metric(lon1,lat1,lon2,lat2);
  if (d==0.0) { *loni=lon1; *lati=lat1; return; }
  if (d==M_PI) { /*antipodes; undefined path*/ *lati=0; *loni=(lon1+lon2)/2.0; return;}
  {
    a=sin((1-fraction)*d)/sin(d) ;
    b=sin(fraction*d)/sin(d) ;
    x = a*cos(lat1)*cos(lon1) +  b*cos(lat2)*cos(lon2) ;
    y = a*cos(lat1)*sin(lon1) +  b*cos(lat2)*sin(lon2) ;
    z = a*sin(lat1)           +  b*sin(lat2) ;
    if (y!=0.0 || x!=0.0) *loni = atan2(y,x); else *loni = lon1; 
    *lati=atan2(z,sqrt(x*x+y*y)) ;
    return;
  }
}

float coursetofrom(float lon2, float lat2, float lon1, float lat1)
{
// from http://williams.best.vwh.net/avform.htm#Crs
  float th;
  //float eps=0.00000003;
  float d;
  float flt;

  //if ( cos(lat1) < eps )   
  if ( cos(lat1) ==0.0 )   
  {
    if (lat1 > 0)
       th= M_PI ;       //  starting from N pole
    else
       th= 0.0 ;         //  starting from S pole
  }
  else
  {
    d = metric(lon1,lat1,lon2,lat2);
    if (sin(d)!=0.0)
    {
      flt = (sin(lat2)-sin(lat1)*cos(d))/(sin(d)*cos(lat1)) ;
      if ( sin(lon1-lon2)<0 ) //    if ( sin(lon2-lon1)<0 )
      {
        if( flt>=-1.0 && flt<=1.0 ) th=acos(flt) ; else th= 0.0;
      }
      else
      {
        if( flt>=-1.0 && flt<=1.0 ) th=M_PIT2-acos(flt) ; else th= 0.0;
      }
    } else th = 0.0; //
  }

  th = M_PIO2 - th ;//  his bearings are east of north and mine are north of east
  th = twopimod(th);

  return th;
}

void rectintermediatepoint(float x1, float y1, float x2, float y2, float fraction, float *ix, float *iy)
{
// returns x,y of point fraction of the way from x1,y1 to x2,y2 on a wrapping rectangle
  float dx1,dx2,dy1,dy2,dret,flt;
  int shortest;
  dx1=x2-x1;
  dy1=y2-y1;
  if ( dy1==0.0 && dx1==0.0 ) { *ix=x1; *iy=y1; return; }
  if(dx1>0.0) 
    dx2=dx1-M_PIT2;
  else
    dx2=dx1+M_PIT2;
  if(dy1>0.0) 
    dy2=dy1-M_PI;
  else
    dy2=dy1+M_PI;
  dret = dx1*dx1 + dy1*dy1; shortest=0;
  flt = dx2*dx2 + dy1*dy1; if(flt<dret) { dret=flt; shortest=1; }
  flt = dx2*dx2 + dy2*dy2; if(flt<dret) { dret=flt; shortest=2; }
  flt = dx1*dx1 + dy2*dy2; if(flt<dret) { dret=flt; shortest=3; }
  switch(shortest)
  {
    case 0:  *ix=x1+fraction*dx1; *iy=y1+fraction*dy1;       break;
    case 1:  *ix=x1+fraction*dx2; *iy=y1+fraction*dy1;       break;
    case 2:  *ix=x1+fraction*dx2; *iy=y1+fraction*dy2;       break;
    case 3:  *ix=x1+fraction*dx1; *iy=y1+fraction*dy2;       break;
  }
  *ix = twopimod( *ix + M_PI) - M_PI ;
  *iy = pimod( *iy + M_PIO2) - M_PIO2 ;
}

float squarecoursetofrom(float x2, float y2, float x1, float y1)
// angle of which direction from x1,y1 to x2,y2
// shortest distance on a wrapping rectangle
{
  float dx1,dx2,dy1,dy2,dret,flt,angret;
  int shortest;
  dx1=x2-x1;
  dy1=y2-y1;
  if ( dy1==0.0 && dx1==0.0 ) return 0.0; 
  if(dx1>0.0) 
    dx2=dx1-M_PIT2;
  else
    dx2=dx1+M_PIT2;
  if(dy1>0.0) 
    dy2=dy1-M_PI;
  else
    dy2=dy1+M_PI;
  dret = dx1*dx1 + dy1*dy1; shortest=0;
  flt = dx2*dx2 + dy1*dy1; if(flt<dret) { dret=flt; shortest=1; }
  flt = dx2*dx2 + dy2*dy2; if(flt<dret) { dret=flt; shortest=2; }
  flt = dx1*dx1 + dy2*dy2; if(flt<dret) { dret=flt; shortest=3; }
  switch(shortest)
  {
    case 0:  angret=atan2(dy1,dx1); break;
    case 1:  angret=atan2(dy1,dx2); break;
    case 2:  angret=atan2(dy2,dx2); break;
    case 3:  angret=atan2(dy2,dx1); break;
  }
  return angret;
}

void walledrectintermediatepoint(float x1, float y1, float x2, float y2, float fraction, float *ix, float *iy)
{
// returns x,y of point fraction of the way from x1,y1 to x2,y2 in flat space
  *ix=x1+fraction*(x2-x1); 
  *iy=y1+fraction*(y2-y1);
}

float walledsquarecoursetofrom(float x2, float y2, float x1, float y1)
// angle of which direction from x1,y1 to x2,y2 in flat space
{
  float dx,dy;
  dx=x2-x1;
  dy=y2-y1;
  if ( dy==0.0 && dx==0.0 ) return 0.0; 
  return atan2(dy,dx);
}

float playermomentflat(float ra)
// gives player saucer moment of inertia from saucer radius(straightened spherical central angle units)
// saucer m is saucer r; saucer I is .5mr^2
{
  return 0.5*ra*ra*ra ; 
}

float playermomentsphere(float ra)
// gives player saucer moment of inertia from saucer radius(in sphere central angle units)
// with the player saucer as a spherical cap
// saucer m is saucer r; saucer I is m1*rs*rs/(1-cos(ra))*( cosra*cosra*cosra/3.0 - cosra + 2.0/3.0 );
//  where rs is the radius of the sphere, 1 here
{
  float cosra;
  cosra=cos(ra);
  //return  ( m1/(1-cos(ra))*( cosra*cosra*cosra/3.0 - cosra + 2.0/3.0 ) ); m1=ra
  return  ( ra/(1-cos(ra))*( cosra*cosra*cosra/3.0 - cosra + 2.0/3.0 ) );
}

float playerorbitmomentsphere(float ra)
// gives player saucer great circle on sphere orbital moment of inertia from saucer radius(in sphere central angle units)
// with the player saucer as a spherical cap
// saucer m is saucer r; saucer orbital I is m1*rs*rs/(1-cos(r1))*( 4.0/3.0 - cos(r1) - cos(r1)*cos(r1)*cos(r1)/3.0 )/2.0;
//  where rs is the radius of the sphere, 1 here
{
  float cosra;
  cosra=cos(ra);
  //return m1*rs*rs/(1-cos(r1))*( 4.0/3.0 - cos(r1) - cos(r1)*cos(r1)*cos(r1)/3.0 )/2.0; m1=ra
  return ( ra/(1-cosra)*( 4.0/3.0 - cosra - cosra*cosra*cosra/3.0 )/2.0 ) ;
}

void bearchang1(float lon, float lat, float vx, float vy, float vz, float *angtra, float *anglook)
{
//  cos new_bearing = ( v dot long.hat) / |v|
//  sin new_bearing = ( v dot lat.hat) / |v|
  float lohx,lohy,lohz; //longitude unit vector
  float lahx,lahy,lahz; //latitude unit vector
  float flt1,flt2,flt3; //temp floats
  //float eps=0.00000000003;

  lohx = -sin(lon);
  lohy = cos(lon);
  lohz = 0.0;

  flt2 = vx*lohx + vy*lohy + vz*lohz;

  lahx = -cos(lon)*sin(lat);
  lahy = -sin(lon)*sin(lat);
  lahz = cos(lat);

  flt3 = vx*lahx + vy*lahy + vz*lahz;
  if (flt3!=0.0 || flt2!=0.0)
  //if (fabs(flt3)>eps || fabs(flt2)>eps)
  {
    flt2 = atan2(flt3,flt2);
    *anglook += flt2 - *angtra;
    *angtra = flt2;
  }
}

//move player method 3
// movement by surfacerun(lon,lat,distance,bearing,&newlon,&newlat)
// solved great circle step
void moveplayerm3(float *lon, float *lat, float vdt, float angtra)
{
  surfacerun(*lon,*lat,angtra,vdt,lon,lat);
}

//move player method 2
// movement by parametric great circle solution equations from
//  http://www.physics.oregonstate.edu/paradigms/Publications/GreatCircles.html
//   equations 8-12
void moveplayerm2(float *arglon, float *arglat, float vdt, float angtra)
{
  double longsp; // starting longitude, primeed coordinates of great circle trajectory
  double longo; //longitude of max latitude of circle
  double latm;  // maximum latitude of circle
  double newlong, newlat;// double eps=0.000003; //
  double flt1,flt2; //temp floats
  double lon = *arglon, lat = *arglat; //local lon and lat, just to not use *var
  int latflip=0;

  if ( cos(angtra)<0.0 ) vdt = -vdt;

  if ( lat<0.0 )  {
    latflip=1; lat=-lat;
    vdt = -vdt;
  }

  latm = acos( fabs( cos(angtra)*cos(lat) ) );

  flt1 = tan(angtra);
  flt2 = sin(lat);
  //  if (fabs(flt1)>eps || fabs(flt2)>eps)
  if (flt1!=0.0 || flt2!=0.0)
  {
    longo = atan2( flt1 , flt2 ) ;
    longo += lon;
  }
  else
    longo = lon; // ?
  if (longo>M_PI) longo-=M_PIT2;
  if (longo<-M_PI) longo+=M_PIT2;

  flt1 = -tan(angtra);
  flt2 = tan(lat) * sqrt( 1.0 + flt1*flt1);
  //  if (fabs(flt1)>eps || fabs(flt2)>eps)
  if (flt1!=0.0 || flt2!=0.0)
    longsp = atan2( flt1 , flt2 );
  else
    longsp = lon; // ?

  newlat = asin( sin(latm) * cos(longsp + vdt) );  // + vdt (units problem? not that I can see.)
  if ( latflip ) { vdt = -vdt; }
  flt1 = tan( longsp + vdt );                     // + vdt
  flt2 = cos( latm );
  //  if (fabs(flt1)>eps || fabs(flt2)>eps)
  if (flt1!=0.0 || flt2!=0.0)
    newlong = atan2( flt1 , flt2 ) + longo;
  else
    newlong = longo; // ?

  if ( latflip ) { newlat = -newlat;  }

  if (newlong>M_PI) newlong-=M_PIT2;
  if (newlong<-M_PI) newlong+=M_PIT2;

  *arglon = newlong;
  *arglat = newlat;
}

void spheremoveplayer(float *lon, float *lat, float v, float *angtra, float *anglook, float anglookv, float dt)
//move by centripetal accelaration (long,lat,speed,bearing,look)
{
  //float r; //radius of sphere
  float x,y,z; //conversion to cartesian coordinates
  float a,ax,ay,az;  //acceleration scalar and vector (centripetal)
  float vx,vy,vz;  //velocity vector
  float newlong, newlat; //
  int trajmaths;

  // for large v, could integrate the following over smaller dt for better accuracy
  x = cos(*lon)*cos(*lat); // *r
  y = sin(*lon)*cos(*lat); // *r
  z = sin(*lat); // *r
  vx = v*(-cos(*angtra)*sin(*lon)-sin(*angtra)*cos(*lon)*sin(*lat));
  vy = v*(cos(*angtra)*cos(*lon)-sin(*angtra)*sin(*lon)*sin(*lat));
  vz = v*(sin(*angtra)*cos(*lat));
  a = v*v; // /r
  ax = -a*x; // /r
  ay = -a*y; // /r
  az = -a*z; // /r
  x += vx*dt + 0.5*ax*dt*dt;
  y += vy*dt + 0.5*ay*dt*dt;
  z += vz*dt + 0.5*az*dt*dt;
  vx += ax*dt;
  vy += ay*dt;
  vz += az*dt;

  trajmaths=trajmath;
  if (trajmaths==2)
  {
    newlong = *lon;  newlat = *lat;
    moveplayerm3(&newlong, &newlat, v*dt, *angtra);
  } 
  if (trajmaths==1)
  {
    if (fabs(*lat)<=v*fabs(dt)) trajmaths=0; //trajmath 1 breaks at the equator
    else
    {
      newlong = *lon;  newlat = *lat;
      moveplayerm2(&newlong, &newlat, v*dt, *angtra);
    }
  } 
  if ( trajmaths==0 )
  {
    if (y!=0.0 || x!=0.0) newlong = atan2(y,x); else newlong = *lon;
    if ( fabs(z)<=1.0 ) newlat = asin(z); else newlat = M_PI*z/fabs(z);
  }

  // curvilinear change of bearing
  bearchang1(newlong,newlat,vx,vy,vz,angtra,anglook);

  *lon = newlong;
  *lat = newlat;

  *lon= twopimod( *lon + M_PI) - M_PI ;
  *anglook+=anglookv * dt ;
  *anglook=twopimod(*anglook);
  *angtra=twopimod(*angtra);
}

void squaremoveplayer(float *x, float *y, float v, float *angtra, float *anglook, float anglookv, float dt)
{
  *x += v * cos(*angtra) * dt ;
  *y += v * sin(*angtra) * dt ;

  *x = twopimod( *x + M_PI) - M_PI ;
  *y = pimod( *y + M_PIO2) - M_PIO2 ;

  *anglook+=anglookv * dt ;
  *anglook=twopimod(*anglook);
  *angtra=twopimod(*angtra);
}

void walledsquaremoveplayer(float *x, float *y, float v, float *angtra, float *anglook, float anglookv, float dt)
{
  *x += v * cos(*angtra) * dt ;
  *y += v * sin(*angtra) * dt ;

  if(*x>M_PI) *x= M_PI; // or --  3.12588
  if(*x<-M_PI) *x= -M_PI;
  if(*y>M_PIO2) *y= M_PIO2;  // or --   1.56215
  if(*y<-M_PIO2) *y= -M_PIO2;

  *anglook+=anglookv * dt ;
  *anglook=twopimod(*anglook);
  *angtra=twopimod(*angtra);
}

void somefriction(int i, int betterfriction, float dt)
{
  float frinc=0.0000001, frictvinc;
  float r1=p1rad*M_PI/Y_RES;
  float r2=p2rad*M_PI/Y_RES;
  frictvinc = frinc*pow(10,i);
  //friction on shots too? probably not  //

  if( betterfriction)
  { // better friction which shows the complication between translation and rotation, 
    // and is numerically integrated
    vector vec;
    {
      if(p1exists && ( p1av!=0.0 || p1v !=0.0 ) ) 
      {
        vec = frictionintegral( frictvinc, p1v, p1av, r1, r1 );
        if (p1av>0.0) { p1av+=vec.z/playermoment(r1)*dt; if(p1av<0.0)p1av=0.0; }
        if (p1av<0.0) { p1av+=vec.z/playermoment(r1)*dt; if(p1av>0.0)p1av=0.0; }
        if (p1v>0.0) 
        {
          if( metric==spheremetric )
            p1v+=vec.y/playerorbitmomentsphere(r1) * dt ;
          else
            p1v+=vec.y/r1 * dt ;
          if (p1v<0.0) p1v=0.0;
        }
      }
      if(p2exists && ( p2av!=0.0 || p2v !=0.0 ) ) 
      {
        vec = frictionintegral( frictvinc, p2v, p2av, r2, r2 );
        if (p2av>0.0) { p2av+=vec.z/playermoment(r2)*dt; if(p2av<0.0)p2av=0.0; }
        if (p2av<0.0) { p2av+=vec.z/playermoment(r2)*dt; if(p2av>0.0)p2av=0.0; }
        if (p2v>0.0) 
        {
          if( metric==spheremetric )
            p2v+=vec.y/playerorbitmomentsphere(r2) * dt ;
          else
            p2v+=vec.y/r2 * dt ;
          if (p2v<0.0) p2v=0.0;
        }
      }
    }
    return;
  }

  if( metric!=spheremetric )
  {
    float p1_rada, p2_rada;
    if(p1av!=0.0) 
    {
      p1_rada = 1.333333/r1;
      if (p1av>0.0) { p1av-=frictvinc*p1_rada*dt; if(p1av<0.0)p1av=0.0; }
      if (p1av<0.0) { p1av+=frictvinc*p1_rada*dt; if(p1av>0.0)p1av=0.0; }
    }
    if(p2av!=0.0) 
    {
      p2_rada = 1.333333/r2;
      if (p2av>0.0) { p2av-=frictvinc*p2_rada*dt; if(p2av<0.0)p2av=0.0; }
      if (p2av<0.0) { p2av+=frictvinc*p2_rada*dt; if(p2av>0.0)p2av=0.0; }
    }
    if (p1v>0.0) 
    {
      p1v-=frictvinc * dt ;
      if (p1v<0.0) p1v=0.0;
    }
    if (p2v>0.0) 
    {
      p2v-=frictvinc * dt ;
      if (p2v<0.0) p2v=0.0;
    }
  }
  else
  { //this T/I is almost identical to that of the flat disc by the way at the cost of more math
    float torque;
    if(p1av!=0.0) 
    {
      //torque = FRINC*m1/2.0/(1-cos(r1))*( r1 - 0.5*sin(2.0*r1) ) ;  m1=r1;
      torque = frictvinc*r1*0.5/(1-cos(r1))*( r1 - 0.5*sin(2.0*r1) ) ;  
      if (p1av>0.0) { p1av-=torque/playermoment(r1)*dt; if(p1av<0.0)p1av=0.0; }
      if (p1av<0.0) { p1av+=torque/playermoment(r1)*dt; if(p1av>0.0)p1av=0.0; }
    }
    if(p2av!=0.0) 
    {
      //m2=r2;
      torque = frictvinc*r2*0.5/(1-cos(r2))*( r2 - 0.5*sin(2.0*r2) ) ;  
      if (p2av>0.0) { p2av-=torque/playermoment(r2)*dt; if(p2av<0.0)p2av=0.0; }
      if (p2av<0.0) { p2av+=torque/playermoment(r2)*dt; if(p2av>0.0)p2av=0.0; }
    }
    if (p1v>0.0) 
    {
      p1v-=frictvinc*r1/playerorbitmomentsphere(r1) * dt ;
      if (p1v<0.0) p1v=0.0;
    }
    if (p2v>0.0) 
    {
      p2v-=frictvinc*r2/playerorbitmomentsphere(r2) * dt ;
      if (p2v<0.0) p2v=0.0;
    }
  }
}

void saveplayers(void)
{
  plrsaverestore[0].exists = p1exists;
  plrsaverestore[0].lon = p1long;
  plrsaverestore[0].lat = p1lat;
  plrsaverestore[0].a = p1a;
  plrsaverestore[0].bear = p1at;
  plrsaverestore[0].v = p1v;
  plrsaverestore[0].av = p1av;
  plrsaverestore[0].rad = p1rad;
  plrsaverestore[0].colc = p1colc;
  plrsaverestore[0].colf = p1colf;
  plrsaverestore[0].shotsfired = p1shotsfired;

  plrsaverestore[1].exists = p2exists;
  plrsaverestore[1].lon = p2long;
  plrsaverestore[1].lat = p2lat;
  plrsaverestore[1].a = p2a;
  plrsaverestore[1].bear = p2at;
  plrsaverestore[1].v = p2v;
  plrsaverestore[1].av = p2av;
  plrsaverestore[1].rad = p2rad;
  plrsaverestore[1].colc = p2colc;
  plrsaverestore[1].colf = p2colf;
  plrsaverestore[1].shotsfired = p2shotsfired;
}

void restoreplayers(void)
{
  p1exists = plrsaverestore[0].exists ;
  p1long = plrsaverestore[0].lon ;
  p1lat = plrsaverestore[0].lat ;
  p1a = plrsaverestore[0].a ;
  p1at = plrsaverestore[0].bear ;
  p1v = plrsaverestore[0].v ;
  p1av = plrsaverestore[0].av ;
  p1rad = plrsaverestore[0].rad ;
  p1colc = plrsaverestore[0].colc ;
  p1colf = plrsaverestore[0].colf ;
  //p1shotsfired = plrsaverestore[0].shotsfired ;

  p2exists = plrsaverestore[1].exists ;
  p2long = plrsaverestore[1].lon ;
  p2lat = plrsaverestore[1].lat ;
  p2a = plrsaverestore[1].a ;
  p2at = plrsaverestore[1].bear ;
  p2v = plrsaverestore[1].v ;
  p2av = plrsaverestore[1].av ;
  p2rad = plrsaverestore[1].rad ;
  p2colc = plrsaverestore[1].colc ;
  p2colf = plrsaverestore[1].colf ;
  //p2shotsfired = plrsaverestore[1].shotsfired ;
}

void init_player2(void)
{
  p2exists=1;
  p2long=-1.571;   //2.0944;
  p2lat=1.0472;
  p2a=5.54; // angle, north of east;
  p2at=5.54;  // angle of travel
  p2v=0.0; //speed
  p2av=0.0; // angular velocity
  //p2colc=255; p2colf=15;
  p2colc=47; p2colf=111;
  p2rad=5.4; // saucer radius in pixels
  //p2shotsfired = 0;
}

void init_player1(void)
{
  p1exists=1;
  p1long=1.571;   //2.0944;
  p1lat=-1.0472;
  p1a=2.3562; // angle, north of east;
  p1at=2.3562;  // angle of travel
  p1v=0.0; // speed
  p1av=0.0; // angular velocity
  //p1colc=31; p1colf=63;  //  p1colc=79; p1colf=207; //  p1colc=239; p1colf=63;
  p1colc=79; p1colf=63;
  p1rad=5.4; // saucer radius in pixels
  //p1shotsfired = 0;
}

void deexist_player2(void)
{
  p2exists=0;
  p2long=0.0;
  p2lat=0.0;
  p2a=0.0;
  p2at=0.0;
  p2v=0.0; 
  p2av=0.0;
  keyp2tl=keyp2tr=0;
}

void deexist_player1(void)
{
  p1exists=0;
  p1long=0.0;
  p1lat=0.0;
  p1a=0.0;
  p1at=0.0;
  p1v=0.0; 
  p1av=0.0;
  keyp1tl=keyp1tr=0;
}

int autocls( clock_t tock, int autoclear )
{
  static int lastclear=0;
  if (lastclear==0) lastclear = tock;
  if (tock<lastclear) lastclear = tock;
  if ( tock > lastclear + (64>>(autoclear-1)) ) //~4 seconds /2^(autoclear-1)
    { clear_screen(); lastclear = tock; return 1;}
  else return 0;
}

// blow grid pixels out from the center of the screen, animation for a second
void gridout()
{
  clock_t start;
  int endrun=0; 
  int x,y,c,xd,yd;
  int i=0; //int k=0;
  start=clock();

  while(!endrun)
  {
    do {
      x=rand()%X_RES;
      y=rand()%Y_RES;
      c=GETPIXS(x,y);
    } while ( c==0 ); //c != gridcol && c != gridcol-BACKPROJDIFF) ;
    {
      PUTPIXS(x,y,0);
      xd = (X_OFFS-x)/14.0 ;
      yd = (Y_OFFS-y)/20.0 ;
      if (xd==0) if( X_OFFS-x > 0 ) xd=1; else xd=-1;
      if (yd==0) if( Y_OFFS-y > 0 ) yd=1; else yd=-1;
      x -= xd ; if (x<0) x=0; if (x>=X_RES) x=X_RESM; 
      y -= yd ; if (y<0) y=0; if (y>=Y_RES) y=Y_RESM; 
      PUTPIXS(x,y,c);  // gridcol);
    }
    i++;
    if(i==800 || gridcol==0) { i=0; if ( clock()>start+17 ) endrun=1; } //k++;}
  }
  //gotoxy(18,12); printf("%d",k);
}

// suck grid pixels to the center of the screen, animation for a second
void gridin()
{
  clock_t start;
  int endrun=0; 
  int x,y,c,xd,yd;
  int i=0; //int k=0;
  start=clock();

  while(!endrun)
  {
    do {
      x=rand()%X_RES;
      y=rand()%Y_RES;
      c=GETPIXS(x,y);
    } while ( c==0 ); //( c != gridcol && c != gridcol-BACKPROJDIFF) ;
    {
      PUTPIXS(x,y,0);
      xd = (X_OFFS-x)/20.0 ;
      yd = (Y_OFFS-y)/14.0 ;
      if (xd==0) if( X_OFFS-x > 0 ) xd=1; else xd=-1;
      if (yd==0) if( Y_OFFS-y > 0 ) yd=1; else yd=-1;
      x += xd ;
      y += yd ;
      if (c==7) { if (rand()%100>31) PUTPIXS(x,y,c); }
      else PUTPIXS(x,y,c); //gridcol);
    }
    i++;
    if(i==800 || gridcol==0) { i=0; if ( clock()>start+19 ) endrun=1; } //k++;}
  }
  //gotoxy(18,12); printf("%d",k);
}

void drawgbazimap(int drawdonterase) // map border for azimuthal projection (around the antipode)
{
#define DRAWGRIDBOAZIMAPR 0.005 // radius around antipode for azimuthal projection border
  float theta=0.0,lon,lat;
  int x,y;
  while (theta<M_PIT2)
  {
    lat= DRAWGRIDBOAZIMAPR * sin(theta);
    lon = M_PI + DRAWGRIDBOAZIMAPR * cos(theta);
    if (lon>M_PI) lon-=M_PIT2;
    projec(lon,lat,&x,&y);
    if(drawdonterase) PUTPIXS(x,y,gridcol-BACKPROJDIFF); else PUTPIXS(x,y,0);
    theta+=M_PI/30;
  }
}

void drawgridbo(int drawdonterase)
{
  float lon,lat;
  int x,y;
  int i,j;
  unsigned char ibits=0, i2=1;
  int erasingsavegridcol;
  float lons[8]={
-3.14159, //3.1415926535897932384626433832795
 3.14159,
-0.00001,
 0.00001,
-1.5708063267,
-1.5707863267,
 1.5707863267,  //this is a fugly method
 1.5708063267   //versus just having short tables for each projection for where its lon and lat borders are
  };
  if(!drawdonterase) { erasingsavegridcol=gridcol; gridcol=0; }
  if ( ( projec!=azimuequiproj && projec!=sphereproj && projec!=twohemiproj && projec!=twohemiproj2 && projec!=perspsphereproj ) 
     ) {ibits|=0x03;}
  if ( projec==sphereproj || projec==perspsphereproj
       || ( projec==mixproj&&( mixprojold==sphereproj||mixprojnew==sphereproj
                             ||mixprojold==perspsphereproj||mixprojnew==perspsphereproj ) )
     ) { ibits|=0xf0; }
  if ( projec==twohemiproj || projec==twohemiproj2 
       || ( projec==mixproj&&(mixprojold==twohemiproj||mixprojold==twohemiproj2) )
       || ( projec==mixproj&&(mixprojnew==twohemiproj||mixprojnew==twohemiproj2) )
     ) { ibits|=0x0f; }
  for(i=0;i<8;i++)
  {
    if (ibits & i2) 
    {
      lon = lons[i];
      for (j=0;j<=150;j++)
      {
        lat = (j/150.0 - .5) * 3.141591;
        projec(lon,lat,&x,&y);
        if (backproj && drawdonterase) PUTPIXS(x,y,gridcol-BACKPROJDIFF);
        else PUTPIXS(x,y,gridcol);
      }
    }
    i2<<=1;
  }
  if ( projec!=azimuequiproj && projec!=sphereproj && projec!=perspsphereproj
       || ( projec==mixproj && ( mixprojold!=azimuequiproj && mixprojold!=sphereproj && mixprojold!=perspsphereproj 
        && mixprojnew!=azimuequiproj && mixprojnew!=sphereproj && mixprojnew!=perspsphereproj  )  )
     )
  for (j=-1;j<=1;j+=2)
  {
    lat = j/2.0* M_PI;
    for (i=0;i<=200;i++)
    {
      lon = (i/100.0 - 1.0) * 3.141591;
      projec(lon,lat,&x,&y);
      if (backproj && drawdonterase) PUTPIXS(x,y,gridcol-BACKPROJDIFF);
      else PUTPIXS(x,y,gridcol);
    }
  }
  if(!drawdonterase) { gridcol=erasingsavegridcol; }
  if ( projec==azimuequiproj ) drawgbazimap(drawdonterase);
}

void drawgrid(int drawdonterase)
{
  float lon,lat;
  int x,y;
  int i,j;
  int erasingsavegridcol;
  float lons[16]={
-3.14159,
-2.6179938779,
-2.0943951023,
-1.5708063267,
-1.5707863267,
-1.0471975511,
-0.5235987755,
-0.00001,
 0.00001,
 3.14159,
 2.6179938779,
 2.0943951023,
 1.5708063267,
 1.5707863267,
 1.0471975511,
 0.5235987755, };
  if(!drawdonterase) { erasingsavegridcol=gridcol; gridcol=0; }
  for (i=0;i<=16;i++)
  {
    lon = lons[i];
    for (j=0;j<=150;j++)
    {
      lat = (j/150.0 - .5) * 3.141591;
      projec(lon,lat,&x,&y);
      if (backproj && drawdonterase) PUTPIXS(x,y,gridcol-BACKPROJDIFF);
      else PUTPIXS(x,y,gridcol);
    }
  }
  for (j=-3;j<=3;j++)
  {
    lat = j/6.0* M_PI;
    for (i=0;i<=200;i++)
    {
      lon = (i/100.0 - 1.0) * 3.141591;
      projec(lon,lat,&x,&y);
      if (backproj && drawdonterase) PUTPIXS(x,y,gridcol-BACKPROJDIFF);
      else PUTPIXS(x,y,gridcol);
    }
  }
  if(!drawdonterase) { gridcol=erasingsavegridcol; }
  if ( projec==azimuequiproj ) drawgbazimap(drawdonterase);
}

void detall(int player)
{
  int i,j;
  float nearest,dist;
  int colf;

  if (player==0)
  {
    colf = p1colf;
  }
  else
  if (player==1)
  {
    colf = p2colf;
  }

  i=0; j=0;
  while ( i<MAX_SHOTS && j<MAX_SHOTS/2 )
  {
    if ( shots[i].owner==player )
    {
      j++;
      if ( !(shots[i].exploding) )
      {
        shots[i].exploding=1;
        shots[i].wipeshottrail=1;
        shots[i].col=colf;
      }
    }
    i++;
  }
}

void detnearest(int player)
{
  int i,j;
  int k=-1;
  float nearest,dist;
  float lon,lat;
  int colf;

  if (player==0)
  {
    lon = p1long;
    lat = p1lat;
    colf = p1colf;
  }
  else
  if (player==1)
  {
    lon = p2long;
    lat = p2lat;
    colf = p2colf;
  }

  i=0; j=0;
  while ( i<MAX_SHOTS && j<MAX_SHOTS/2 )
  {
    if ( shots[i].owner==player )
    {
      j++;
      if ( !(shots[i].exploding) && !(shots[i].fresh) )
      {
        if (k==-1) { nearest = metric( shots[i].lon, shots[i].lat, lon, lat); k=i; }
        else
        {
          dist = metric( shots[i].lon, shots[i].lat, lon, lat);
          if ( dist<nearest ) { nearest=dist; k=i; }
        }
      }
    }
    i++;
  }
  if (k!=-1)
  {
    shots[k].exploding=1;
    shots[k].wipeshottrail=1;
    shots[k].col=colf;
  }
}

void shoot(int player)
// pv: player speed, pat: player angle of travel, pa: player look angle
// mv: muzzle speed, sv: shot speed, sat: shot angle of travel
// shots are color player colc; they will explode in player colf or white?
// masses : player mass is prad, shot mass is .5
{
  int shotsinair, i, ilook, iselect=-1;
  float lon,lat,pa,*pat,*pv,prad;
  int col,colf,*shotsfired;
  float mv=muzzlespeed;
  float sv,sat,flt2,flt3,mvrecoil;

  if (player==0)
  {
    lon = p1long;
    lat = p1lat;
    pa = p1a;
    pat = &p1at;
    pv = &p1v; //speed
    prad = p1rad; // mass = player saucer radius 
    col = p1colc;
    colf = p1colf;
    shotsfired = &p1shotsfired;
  }
  else
  if (player==1)
  {
    lon = p2long;
    lat = p2lat;
    pa = p2a;
    pat = &p2at;
    pv = &p2v; //speed
    prad = p2rad; // mass = player saucer radius 
    col = p2colc;
    colf = p2colf;
    shotsfired = &p2shotsfired;
  }
  if(metric==spheremetric) prad=playerorbitmomentsphere(prad*M_PI/Y_RES)/M_PI*Y_RES; // unit sphere radius simplifies this treatment
  shotsinair=0;
  ilook=-1;
  for( i=0; i<MAX_SHOTS; i++ )
  {
    if (shots[i].owner==-1) iselect=i;  //select free shot slot
    else
    if (shots[i].owner==player)
    {
      shotsinair++;
      if ( shots[i].exploding==0 && shots[i].fresh==0 ) //will detonate oldest shot if none free
      {
        if ( ilook==-1 )
          ilook=i;
        else
          if (shots[i].serial<shots[ilook].serial) ilook=i;
      }
    }
  }
  if ( shotsinair<MAX_SHOTS/2 && iselect>=0 ) //
  { //shoot
    *shotsfired=(*shotsfired)+1;
    shots[iselect].owner=player;
    shots[iselect].serial=*shotsfired;
    shots[iselect].lon = lon;
    shots[iselect].lat = lat;
    shots[iselect].col=col;
    shots[iselect].rad=.5;
    shots[iselect].fresh=1;
    shots[iselect].exploding=0;
    shots[iselect].stale=0;
    shots[iselect].wipeshottrail=0;
    shots[iselect].oldindex=0;
    for (i=0;i<SHOTDRAWMEM;i++)
    {
      shots[iselect].oldx[i]=-1;
      shots[iselect].oldy[i]=-1;
    }
    // conserved momentum while the player creates the mass for the shot
    *pv= prad/(prad+0.5) * *pv;
    // firing
    flt3=(*pv*cos(*pat)+mv*cos(pa));
    flt2=(*pv*sin(*pat)+mv*sin(pa));
    if(flt2!=0.0||flt3!=0.0) shots[iselect].bear = sat = atan2(flt2,flt3);
    shots[iselect].v = sv = sqrt(flt3*flt3+flt2*flt2);
    // player recoil from firing
    mvrecoil = mv * .5 / prad ;
    flt3=*pv*cos(*pat) - mvrecoil*cos(pa);
    flt2=*pv*sin(*pat) - mvrecoil*sin(pa);
    if(flt2!=0.0||flt3!=0.0) *pat=atan2(flt2,flt3);
    *pv=sqrt(flt3*flt3+flt2*flt2);
  }
  else if (ilook!=-1)
  { //detonate oldest shot
    shots[ilook].wipeshottrail=1;
    shots[ilook].exploding=1;
    shots[ilook].col=colf;
  }
}

void moveshots(float dt)
{
  int i;
  float flt1,flt2;

  flt1=0.0;   flt2=0.0;
  for (i=0;i<MAX_SHOTS;i++)
  {
    if ( (shots[i].owner)!=-1)
    {
      moveplayer(&(shots[i].lon),&(shots[i].lat),shots[i].v,&(shots[i].bear),&flt1,flt2,dt);
      if ( (shots[i].exploding)==1)
      {
        shots[i].rad = shots[i].rad + SHOTEXPLVEL*dt;
        shots[i].v -= shots[i].v * 0.0037 * dt;  //drag
        flt1=0.00001; // friction, 
        if (shots[i].v>0.0) shots[i].v-=flt1;
        if (shots[i].v<0.0) shots[i].v=0.0;
        if ( shots[i].rad>shotexplrad )
        {
          shots[i].owner=-1;
        }
      }
    }
  }
}

void flushshotstreaks(void)
{
  int i,j;
  for (i=0;i<MAX_SHOTS;i++)
  {
    if (shots[i].owner!=-1)
      for (j=0;j<SHOTDRAWMEM;j++)
      {
        shots[i].oldx[j]=-1;
        shots[i].oldy[j]=-1;
      }
  }
}

void drawshots(void)
{
  int i;

  for (i=0;i<MAX_SHOTS;i++)
  {
    if (shots[i].owner!=-1)
    {
      //drawobj(shots[i].lon, shots[i].lat, shots[i].bear, shots[i].rad, shots[i].col, shots[i].col);
      drawshot(i,0); // shot i, not erasing
    }
  }
}

void eraseshots(void)
{
  int i;

  for (i=0;i<MAX_SHOTS;i++)
  {
    if (shots[i].owner!=-1)
    {
      //drawobj(shots[i].lon, shots[i].lat, shots[i].bear, shots[i].rad, 0, 0);
      if ( shots[i].wipeshottrail ) { eraseshottrailall(i); shots[i].wipeshottrail=0; }
      if ( (shots[i].exploding)==1)
      {
        drawshot(i,1); // shot i, erasing
      }
      else eraseshottraillast(i);
    }
  }
}

void initshots(void)
{
  int i;

  for (i=0;i<MAX_SHOTS;i++)
  {
    shots[i].owner=-1;
  }
}

void emergeshots(void)
{
  int i;
  int owner;
  float dist;
  float slon,slat,srad;

  for (i=0;i<MAX_SHOTS;i++)
  {
    owner = shots[i].owner;
    if (owner!=-1)
    {
      slon = shots[i].lon;
      slat = shots[i].lat;
      srad = shots[i].rad;
      if  ( shots[i].fresh )
      {
        if (owner==0)
        {
          dist = metric( p1long,p1lat, slon,slat);
          if ( dist > (p1rad+srad)*M_PI/Y_RES || !p1exists)
            shots[i].fresh = 0;
        }
        if (owner==1)
        {
          dist = metric( p2long,p2lat, slon,slat);
          if ( dist > (p2rad+srad)*M_PI/Y_RES || !p2exists)
            shots[i].fresh = 0;
        }
      }
    }
  }
}

void playerwallpushout(float th, float indist, float *plon, float *plat)
{
  *plon = *plon + indist*cos(th) ;
  *plat = *plat + indist*sin(th) ;
}

void playerwallbouncel(float th, float prad, float *pv, float *pat, float *pav)
{
  float uf=.1; // coefficient of friction for transverse impulse
  float v1it, v1ii; // v1[initial or final][ (theta or theta-transverse) ]
  float v1ft, v1fi; // v1[initial or final][ (theta or i) ]
  float w1i, w1f; // angular velocity, w1[initial or final]
  float r1; //saucer radius
  float m1; //mass
  float i1; //moment of inertia
  float v1pii; //initial transverse velocity, point of contact
  float d1pii; //initial transverse relative velocity, point of contact
  float td1; // tilde d1pii; normed to a sign, or reduced to a smaller factor 
  int overfric1; // when friction force has been applied unphysically far

  r1 = prad*M_PI/Y_RES; 

  v1it = *pv * cos ( *pat - th ) ;
  v1ii = *pv * sin ( *pat - th ) ;

  m1 = r1;
  i1 = playermoment(r1);

  w1i = *pav;

  v1ft = -v1it ;

  v1pii = v1ii + r1*w1i ;
  d1pii = v1pii ;  

  if (d1pii==0.0) { td1=0.0; }
  else
  {
    if (d1pii>0.0) { td1=1.0; }
    else { td1=-1.0; }
  }

  v1fi = v1ii  -  uf * fabs( v1ft - v1it ) * td1 ;
  w1f  = w1i   -  r1 / i1 * uf * m1 * fabs( v1ft - v1it ) * td1 ;

  if ( ( v1fi + r1*w1f ) * td1 < 0.0 ) overfric1 = 1; else overfric1 = 0;

  if ( overfric1 ) 
  {
    td1  = d1pii / ( uf * fabs( v1ft - v1it ) * (1.0 + m1*r1*r1/i1) ) ;
    v1fi = v1ii  -  uf * fabs( v1ft - v1it ) * td1 ;
    w1f  = w1i   -  r1 / i1 * uf * m1 * fabs( v1ft - v1it ) * td1 ;
  }

  if ( uf!=0.0 && m1*v1fi*v1fi+i1*w1f*w1f>m1*v1ii*v1ii+i1*w1i*w1i )
  { //I don't think this ever happens.
    float tildeke,a,b;
    b  = -2.0*m1*v1ii*(uf * fabs( v1ft - v1it ) * td1) -2.0*i1*w1i*(r1 / i1 * uf * m1 * fabs( v1ft - v1it ) * td1) ;
    a  = m1*pow(uf * fabs( v1ft - v1it ) * td1,2) + i1*pow(r1 / i1 * uf * m1 * fabs( v1ft - v1it ) * td1,2) ;
    tildeke = -b/a;
    if(tildeke<0.0)tildeke=0.0;
    v1fi = v1ii  -  uf * fabs( v1ft - v1it ) * td1 *tildeke ;
    w1f  = w1i   -  r1 / i1 * uf * m1 * fabs( v1ft - v1it ) * td1 *tildeke;
  }

  *pav = w1f;

  *pv = sqrt( v1ft*v1ft + v1fi*v1fi ) ;
  if ( v1fi!=0.0 || v1ft!=0.0 ) *pat = twopimod( atan2(v1fi,v1ft) + th );
}

void playerwallbounce(float th, float prad, float *pv, float *pat, float *pav)
{
  float v1it, v1ii; // v1[initial or final][ (theta or theta-transverse) ]
  float v1ft, v1fi; // v1[initial or final][ (theta or i) ]

  v1it = *pv * cos ( *pat - th ) ;
  v1ii = *pv * sin ( *pat - th ) ;

  v1ft = -v1it ;
  v1fi = v1ii ;

  *pv = sqrt( v1ft*v1ft + v1fi*v1fi ) ;
  if ( v1fi!=0.0 || v1ft!=0.0 ) *pat = twopimod( atan2(v1fi,v1ft) + th );
}

int wallcolliding(float lon, float lat, float prad, float *th, float *dist)
{
  float prada=prad*M_PI/Y_RES;
  float diff;
  int colliding=0;
  *dist= M_PI - (lon+prada) ; // +PI x-wall
  if(*dist<=0.0) { *th = 0.0; colliding=1; }
  diff= M_PI + (lon-prada);  // -PI x-wall
  if(diff<=0.0) { if(colliding==0||diff>*dist){ *dist=diff; *th = M_PI; } colliding=1; }
  diff= M_PIO2 - (lat+prada);  // +PI/2 y-wall
  if(diff<=0.0) { if(colliding==0||diff>*dist){ *dist=diff; *th = M_PIO2; } colliding=1; }
  diff= M_PIO2 + (lat-prada);  // -PI/2 y-wall
  if(diff<=0.0) { if(colliding==0||diff>*dist){ *dist=diff; *th = 3*M_PIO2; } colliding=1; }
  return colliding;
}

int p1p2wallcollisions(void)
{
  static int p1colliding=0, p2colliding=0;
  float p1collth, p2collth, p1in, p2in;
  int wasonwall=0;

  if (!p1colliding)
  {
    if ( ( p1colliding=wallcolliding(p1long,p1lat,p1rad,&p1collth,&p1in) ) ) 
    {
      float w1i=p1av;
      playerwallbounces(p1collth,p1rad,&p1v,&p1at,&p1av);
      // when player is manually turning, let the angular velocity change from this collision show 
      // by deactivating their turn command
      if( p1turnl && w1i!=0.0 && fabs(p1av/w1i-1.0)>=AVDE ) { p1turnl=3; } 
      if( p1turnr && w1i!=0.0 && fabs(p1av/w1i-1.0)>=AVDE ) { p1turnr=3; } 
      wasonwall=1;
    }
  }
  else
  {
    if ( wallcolliding(p1long,p1lat,p1rad,&p1collth,&p1in) )
    {
      playerwallpushout(p1collth,p1in,&p1long,&p1lat);
      p1colliding=0;
      wasonwall=1;
    }
    else
      p1colliding=0;
  }
  if (!p2colliding)
  {
    if ( ( p2colliding=wallcolliding(p2long,p2lat,p2rad,&p2collth,&p2in) ) ) 
    {
      float w2i=p2av;
      playerwallbounces(p2collth,p2rad,&p2v,&p2at,&p2av);
      if( p2turnl && w2i!=0.0 && fabs(p2av/w2i-1.0)>=AVDE ) { p2turnl=3; }
      if( p2turnr && w2i!=0.0 && fabs(p2av/w2i-1.0)>=AVDE ) { p2turnr=3; }
      wasonwall=1; 
    }
  }
  else
  {
    if ( wallcolliding(p2long,p2lat,p2rad,&p2collth,&p2in) )
    {
      playerwallpushout(p2collth,p2in,&p2long,&p2lat);
      p2colliding=0;
      wasonwall=1;
    }
    else
      p2colliding=0;
  }
  return wasonwall;
}

void shotwallcollisions()
{
  int i;
  float th,d;
  float srad, slon, slat;

  for (i=0;i<MAX_SHOTS;i++)
  {
    if (shots[i].owner!=-1 && !shots[i].exploding)
    {
      srad = shots[i].rad;
      slon = shots[i].lon;
      slat = shots[i].lat;
      if ( wallcolliding(slon,slat,srad,&th,&d) )
      {
        if ( shots[i].owner==0 ) shots[i].col=p1colf;
        if ( shots[i].owner==1 ) shots[i].col=p2colf;
        shots[i].wipeshottrail=1;
        shots[i].exploding=1;
      }
    }
  }
}

void sphereabsorbknockturn(float *av1, float r1, float dr1, float dist, float m2, float v2, float th)
//player saucer absorbing a shot's angular momentum, plus reduced spin with saucer growth
{
  float i1i, i1f;
  i1i = playermoment(r1);
  i1f = playermoment(r1+dr1);
  *av1 = ( i1i * *av1 + sin(dist)*m2*v2*sin(th) ) /i1f ;
}

void flatabsorbknockturn(float *av1, float r1, float dr1, float dist, float m2, float v2, float th)
//player saucer absorbing a shot's angular momentum, plus reduced spin with saucer growth
{
  float i1i, i1f;
  i1i = playermoment(r1);
  i1f = playermoment(r1+dr1);
  *av1 = ( i1i * *av1 + dist*m2*v2*sin(th) ) /i1f ;
}

void shotplrcollisions(int shotkollangular, float dt)
// shot damage increases player size; shot imparts momentum
{
  int i;
  int colf;
  int owner;
  float dist1,dist2;
  float vl,vf,vsl,vsf,vst,vsi ;
  float th=0.0;
  float fraction,loni,lati;
  float srad, sv, sang, slon, slat;
  float dpr; //delta player radius, in pixel units
  float sev; //shot explosion speed
  int staleit;
  int plrloopi;
  float plon, plat, prad, pm, pv, pat, pav, dist, pmdpr;

  for (i=0;i<MAX_SHOTS;i++)
  {
    owner = shots[i].owner;
    if (owner!=-1)
    {
      srad = shots[i].rad;
      sv = shots[i].v;
      sang = shots[i].bear;
      staleit = 0 ;
      slon = shots[i].lon;
      slat = shots[i].lat;
      sev = SHOTEXPLVEL*M_PI/Y_RES;
      dist1 = metric( p1long,p1lat, slon,slat);
      dist2 = metric( p2long,p2lat, slon,slat);

  for ( plrloopi=0; plrloopi<=1; plrloopi++ )
  {
      float w1i, w2i;
      if (plrloopi==0) { plon=p1long; plat=p1lat; prad=p1rad; pv=p1v; pat=p1at; pav=p1av; dist=dist1; w1i=p1av; if(!p1exists)continue; }
      if (plrloopi==1) { plon=p2long; plat=p2lat; prad=p2rad; pv=p2v; pat=p2at; pav=p2av; dist=dist2; w2i=p2av; if(!p2exists)continue; }
      if ( !shots[i].fresh && dist < (prad+srad)*M_PI/Y_RES )
      { 
        pm = prad;
        if(metric==spheremetric) pm=playerorbitmomentsphere(pm*M_PI/Y_RES)/M_PI*Y_RES; // unit sphere radius simplifies this
        vl = pv * cos(pat) ;
        vf = pv * sin(pat) ;
        if(metric!=spheremetric)
        {
          vsl = sv * cos( sang ) ;
          vsf = sv * sin( sang ) ;
        }
        fraction = ( prad - ( prad + srad - dist*Y_RES/M_PI ) *0.5 ) / dist *M_PI/Y_RES ;
        intermediatepoint(plon, plat, slon, slat, fraction, &loni, &lati);
        th = angtofrom(loni, lati, plon, plat) ;
        if ( !(shots[i].exploding) )
        {
          dpr = .73 ;
          if(metric==spheremetric)
          {
            vector mergv; //velocity vector of player after merged with shot
            pmdpr = playerorbitmomentsphere((prad+dpr)*M_PI/Y_RES)/M_PI*Y_RES ;
            mergv=wcrossr(vecttimes(1.0/pmdpr,vectsum( vecttimes(pm,rcrossv(plon,plat,pv,pat)),vecttimes(0.5,rcrossv(slon,slat,sv,sang)) ) ),plon,plat);
            bearchang1(plon,plat,mergv.x,mergv.y,mergv.z,&pat,&pmdpr); // pat = atan( (mergv dot lat hat) / (mergv dot lon hat) ), pmdpr used as dummy
            pv=sqrt(vectdot(mergv,mergv));
          }
          else 
          {
            pmdpr = pm + dpr ;
            vl = ( pm * vl + 0.5 * vsl ) / pmdpr ; 
            vf = ( pm * vf + 0.5 * vsf ) / pmdpr ;
            pv = sqrt (vl*vl + vf*vf) ;
            if ( vl != 0.0 || vf != 0.0 ) pat = atan2(vf,vl) ;
          }
          if(shotkollangular) absorbknockturn(&pav, prad*M_PI/Y_RES, dpr*M_PI/Y_RES, dist, 0.5*M_PI/Y_RES, sv, sang-th);
          prad+=dpr;
          eraseshottrailall(i);
          shots[i].owner=-1;
        }
        else  // exploding shot
        {
          vst = sv * cos ( sang - th ) - sev ;
          vsi = sv * sin ( sang - th ) ;
          if ( !(shots[i].stale) )
          { dpr = (0.73 - srad/shotexplrad*.33); loni=1.0; staleit = 1; }
          else { dpr=0.0; loni=.2/srad*dt;} //0.0;}//1.0/18.0; }
          if(metric==spheremetric) pmdpr = playerorbitmomentsphere((prad+dpr)*M_PI/Y_RES)/M_PI*Y_RES ;
          else pmdpr = pm + dpr ;
          vl = ( pm*pv*cos(pat-th) + 0.5*(1.1-srad/shotexplrad)*vst*loni ) / pmdpr; 
          vf = ( pm*pv*sin(pat-th) + 0.5*(1.1-srad/shotexplrad)*0.1*vsi*loni ) / pmdpr; 
          pv = sqrt( vl*vl + vf*vf ) ; //misnomers: vl and vf here are vt and vi (theta and theta-transverse)
          if ( vl != 0.0 || vf != 0.0 ) pat = twopimod( atan2(vf,vl) + th );
          if( shotkollangular && !(shots[i].stale) )
          {
            if ( vsi != 0.0 || vst != 0.0 ) th = atan2(vsi,vst) ;
            vst = sqrt( vst*vst + vsi*vsi ) ;
            absorbknockturn(&pav, prad*M_PI/Y_RES, dpr*M_PI/Y_RES, dist, 0.5*M_PI/Y_RES*dpr/0.73, vst, th);
          }
          prad+=dpr;
        }
      } 
      if (plrloopi==0) 
      { 
        p1long=plon; p1lat=plat; p1rad=prad; p1v=pv; p1at=pat; p1av=pav;
        if( p1turnl && w1i!=0.0 && fabs(p1av/w1i-1.0)>=AVDE ) { p1turnl=3; } 
        if( p1turnr && w1i!=0.0 && fabs(p1av/w1i-1.0)>=AVDE ) { p1turnr=3; } 
      } 
      if (plrloopi==1) 
      { 
        p2long=plon; p2lat=plat; p2rad=prad; p2v=pv; p2at=pat; p2av=pav; 
        if( p2turnl && w2i!=0.0 && fabs(p2av/w2i-1.0)>=AVDE ) { p2turnl=3; }
        if( p2turnr && w2i!=0.0 && fabs(p2av/w2i-1.0)>=AVDE ) { p2turnr=3; }
      } 
  }
      if (staleit) shots[i].stale=1 ;
    }
  }
}

void p1p2pushout4(float dist)
{
  float th1,th2; //collision line bearings, theta
  float halfovrlp,midovrlp1,midovrlp2;
  float p12radii;
  float loni,lati;
  float fraction;

  p12radii = (p1rad+p2rad)*M_PI/Y_RES;
  halfovrlp = ( p12radii - dist ) *0.5 ; 
  midovrlp1 = ( p12radii - dist )*p2rad/(p1rad+p2rad) ; 
  midovrlp2 = p12radii - dist - midovrlp1 ; 

  fraction = ( p1rad*M_PI/Y_RES - halfovrlp ) / dist ;
  intermediatepoint(p1long, p1lat, p2long, p2lat, fraction, &loni, &lati);
  th1 = M_PI + angtofrom(loni, lati, p1long, p1lat) ;
  th2 = M_PI + angtofrom(loni, lati, p2long, p2lat) ;

  surfacerun(p1long,p1lat,th1,midovrlp1,&p1long,&p1lat);
  surfacerun(p2long,p2lat,th2,midovrlp2,&p2long,&p2lat);
}

float contactfrictionorbitalmomentarm(float r1)
{
//  return 0.5*(1-cos(r1))*(1-cos(r1))/(1+cos(r1)) ; //this was misguided
  return 0.5*(1+cos(r1)) ;  // moment arm from sphere center to cap center of mass 
}

void p1p2bounce_sphere(float dist)
// collision response for spherical shell caps on sphere surface, with transverse velocity friction spin
{
  float th1,th2; //collision line bearings, theta
  float uf=uf_p2p; // coefficient of friction for transverse impulse
  float v1it, v1ii, v2it, v2ii; // v[player 1 or 2][initial or final][ (theta or theta-transverse) ]
  float v1ft, v1fi, v2ft, v2fi; // v[1 or 2]       [i or f]          [ (t or i) ]
  float w1i, w2i; // angular velocity, w[player 1 or 2][initial or final]
  float w1f, w2f;
  float r1,r2; //saucer radii
  float m1,m2; //masses
  float i1,i2; //spin moments of inertia
  float i1orbit, i2orbit; //orbital moments of inertia
  float v1pii, v2pii; //initial transverse velocities, point of contact
  float d1pii, d2pii; //initial transverse relative velocities, point of contact
  float td1, td2; // tilde d1pii and d2pii; normed to a sign, or reduced to a smaller factor 
  int overfric1, overfric2; // when friction force has been applied unphysically far
  float loni,lati;
  float p12radii, halfovrlp;
  float fraction;

  p12radii = (p1rad+p2rad)*M_PI/Y_RES;
  halfovrlp = ( p12radii - dist ) *0.5 ; 
  fraction = ( p1rad*M_PI/Y_RES - halfovrlp ) / dist ;
  intermediatepoint(p1long, p1lat, p2long, p2lat, fraction, &loni, &lati);

  th1 = angtofrom(loni, lati, p1long, p1lat) ;
  th2 = M_PI + angtofrom(loni, lati, p2long, p2lat) ; 
  //{ gotoxy(10,1); printf("%f, %f",th1,th2); }

  v1it = p1v * cos ( p1at - th1 ) ;
  v2it = p2v * cos ( p2at - th2 ) ; if (v1it-v2it<0.0) return;
  v1ii = p1v * sin ( p1at - th1 ) ;
  v2ii = p2v * sin ( p2at - th2 ) ;

  r1 = p1rad*M_PI/Y_RES; 
  r2 = p2rad*M_PI/Y_RES; 
  m1 = r1;
  m2 = r2;
  i1 = playermoment(r1);
  i2 = playermoment(r2);
  i1orbit = playerorbitmomentsphere(r1);
  i2orbit = playerorbitmomentsphere(r2);

  w1i = p1av;
  w2i = p2av;

  v1ft = ( (i1orbit-i2orbit) * v1it  +  2.0 * i2orbit * v2it ) / ( i1orbit + i2orbit ) ;
  v2ft = ( (i2orbit-i1orbit) * v2it  +  2.0 * i1orbit * v1it ) / ( i1orbit + i2orbit ) ;

  v1pii = v1ii*cos(r1) + sin(r1)*w1i ;
  v2pii = v2ii*cos(r2) - sin(r2)*w2i ;
  d1pii = v1pii - v2pii ;  // v1ii + r1*w1i - v2ii + r2*w2i;
  d2pii = -d1pii;

  if (d1pii==0.0) { td1=0.0; td2=0.0; }
  else
  {
    if (d1pii>0.0) { td1=1.0; td2=-1.0; }
    else { td1=-1.0; td2=1.0; }
  }

  v1fi = v1ii  -  contactfrictionorbitalmomentarm(r1) * uf * fabs( v1ft - v1it ) * td1 ;
  v2fi = v2ii  -  contactfrictionorbitalmomentarm(r2) * uf * fabs( v2ft - v2it ) * td2 ;
  w1f  = w1i   -  sin(r1) / i1 * uf * i1orbit * fabs( v1ft - v1it ) * td1 ;
  w2f  = w2i   +  sin(r2) / i2 * uf * i2orbit * fabs( v2ft - v2it ) * td2 ;

  if ( ( v1fi*cos(r1) + sin(r1)*w1f - v2pii ) * td1 < 0.0 ) overfric1 = 1; else overfric1 = 0;
  if ( ( v2fi*cos(r2) - sin(r2)*w2f - v1pii ) * td2 < 0.0 ) overfric2 = 1; else overfric2 = 0;

  if( overfric1 || overfric2 )
  {
    if ( overfric1 ) 
      td1  = d1pii / ( uf * fabs( v1ft - v1it ) * ( contactfrictionorbitalmomentarm(r1)*cos(r1) + i1orbit*sin(r1)*sin(r1)/i1) ) ;
    if ( overfric2 )
      td2  = d2pii / ( uf * fabs( v2ft - v2it ) * ( contactfrictionorbitalmomentarm(r2)*cos(r2) + i2orbit*sin(r2)*sin(r2)/i2) ) ;
    if ( fabs(td1)<fabs(td2) ) td2=fabs(td1)*td2/fabs(td2) ;
    if ( fabs(td2)<fabs(td1) ) td1=fabs(td2)*td1/fabs(td1) ;
    v1fi = v1ii  -  contactfrictionorbitalmomentarm(r1) * uf * fabs( v1ft - v1it ) * td1 ;
    w1f  = w1i   -  sin(r1) / i1 * uf * i1orbit * fabs( v1ft - v1it ) * td1 ;
    v2fi = v2ii  -  contactfrictionorbitalmomentarm(r2) * uf * fabs( v2ft - v2it ) * td2 ;
    w2f  = w2i   +  sin(r2) / i2 * uf * i2orbit * fabs( v2ft - v2it ) * td2 ;
  }

  if ( uf!=0.0 && i1orbit*v1fi*v1fi+i1*w1f*w1f+i2orbit*v2fi*v2fi+i2*w2f*w2f>i1orbit*v1ii*v1ii+i1*w1i*w1i+i2orbit*v2ii*v2ii+i2*w2i*w2i )
  {
    float tildeke,a,b;
    b  = -2.0*i1orbit*v1ii*(contactfrictionorbitalmomentarm(r1) * uf * fabs( v1ft - v1it ) * td1) -2.0*i1*w1i*(sin(r1) / i1 * uf * i1orbit * fabs( v1ft - v1it ) * td1) ;
    b += -2.0*i2orbit*v2ii*(contactfrictionorbitalmomentarm(r2) * uf * fabs( v2ft - v2it ) * td2) +2.0*i2*w2i*(sin(r2) / i2 * uf * i2orbit * fabs( v2ft - v2it ) * td2) ;
    a  = i1orbit*pow(contactfrictionorbitalmomentarm(r1) * uf * fabs( v1ft - v1it ) * td1,2) + i1*pow(sin(r1) / i1 * uf * i1orbit * fabs( v1ft - v1it ) * td1,2) ;
    a += i2orbit*pow(contactfrictionorbitalmomentarm(r2) * uf * fabs( v2ft - v2it ) * td2,2) + i2*pow(sin(r2) / i2 * uf * i2orbit * fabs( v2ft - v2it ) * td2,2) ;
    tildeke = -b/a;
    if(tildeke<0.0)tildeke=0.0;
    v1fi = v1ii  -  contactfrictionorbitalmomentarm(r1) * uf * fabs( v1ft - v1it ) * td1 *tildeke;
    w1f  = w1i   -  sin(r1) / i1 * uf * i1orbit * fabs( v1ft - v1it ) * td1 *tildeke;
    v2fi = v2ii  -  contactfrictionorbitalmomentarm(r2) * uf * fabs( v2ft - v2it ) * td2 *tildeke;
    w2f  = w2i   +  sin(r2) / i2 * uf * i2orbit * fabs( v2ft - v2it ) * td2 *tildeke;
  }

  // when player is manually turning, let the angular velocity change from this collision show 
  // by deactivating their turn command
  if( p1turnl && w1i!=0.0 && fabs(w1f/w1i-1.0)>=AVDE ) { p1turnl=3; } 
  if( p1turnr && w1i!=0.0 && fabs(w1f/w1i-1.0)>=AVDE ) { p1turnr=3; }
  if( p2turnl && w2i!=0.0 && fabs(w2f/w2i-1.0)>=AVDE ) { p2turnl=3; }
  if( p2turnr && w2i!=0.0 && fabs(w2f/w2i-1.0)>=AVDE ) { p2turnr=3; }

  p1av = w1f;
  p2av = w2f;

  p1v = sqrt( v1ft*v1ft + v1fi*v1fi ) ;
  if ( v1fi!=0.0 || v1ft!=0.0 ) p1at = twopimod( atan2(v1fi,v1ft) + th1 );
  p2v = sqrt( v2ft*v2ft + v2fi*v2fi ) ;
  if ( v2fi!=0.0 || v2ft!=0.0 ) p2at = twopimod( atan2(v2fi,v2ft) + th2 );
}

void p1p2bounce_flat(float dist)
// collision response for circles in 2-d, with transverse velocity friction spin
{
  float th1,th2; //collision line bearings, theta
  float uf=uf_p2p; // coefficient of friction for transverse impulse
  float v1it, v1ii, v2it, v2ii; // v[player 1 or 2][initial or final][ (theta or theta-transverse) ]
  float v1ft, v1fi, v2ft, v2fi; // v[1 or 2]       [i or f]          [ (t or i) ]
  float w1i, w2i; // angular velocity, w[player 1 or 2][initial or final]
  float w1f, w2f;
  float r1,r2; //saucer radii
  float m1,m2; //masses
  float i1,i2; //moments of inertia
  float v1pii, v2pii; //initial transverse velocities, point of contact
  float d1pii, d2pii; //initial transverse relative velocities, point of contact
  float td1, td2; // tilde d1pii and d2pii; normed to a sign, or reduced to a smaller factor 
  int overfric1, overfric2; // when friction force has been applied unphysically far
  float loni,lati;
  float p12radii, halfovrlp;
  float fraction;

  p12radii = (p1rad+p2rad)*M_PI/Y_RES;
  halfovrlp = ( p12radii - dist ) *0.5 ; 
  fraction = ( p1rad*M_PI/Y_RES - halfovrlp ) / dist ;
  intermediatepoint(p1long, p1lat, p2long, p2lat, fraction, &loni, &lati);

  th1 = angtofrom(loni, lati, p1long, p1lat) ;
  th2 = M_PI + angtofrom(loni, lati, p2long, p2lat) ; 
  // th2==th1, theta code here is from when sphere and flat surfaces shared this routine
  //{ gotoxy(10,1); printf("%f, %f",th1,th2); }

  v1it = p1v * cos ( p1at - th1 ) ;
  v2it = p2v * cos ( p2at - th2 ) ; if (v1it-v2it<0.0) return;
  v1ii = p1v * sin ( p1at - th1 ) ;
  v2ii = p2v * sin ( p2at - th2 ) ;

  r1 = p1rad*M_PI/Y_RES; 
  r2 = p2rad*M_PI/Y_RES; 
  m1 = r1;
  m2 = r2;
  i1 = playermoment(r1);
  i2 = playermoment(r2);

  w1i = p1av;
  w2i = p2av;

  v1ft = ( (m1-m2) * v1it  +  2.0 * m2 * v2it ) / ( m1 + m2 ) ;
  v2ft = ( (m2-m1) * v2it  +  2.0 * m1 * v1it ) / ( m1 + m2 ) ;

  v1pii = v1ii + r1*w1i ;
  v2pii = v2ii - r2*w2i ;
  d1pii = v1pii - v2pii ;  // v1ii + r1*w1i - v2ii + r2*w2i;
  d2pii = -d1pii;

  if (d1pii==0.0) { td1=0.0; td2=0.0; }
  else
  {
    if (d1pii>0.0) { td1=1.0; td2=-1.0; }
    else { td1=-1.0; td2=1.0; }
  }

  v1fi = v1ii  -  uf * fabs( v1ft - v1it ) * td1 ;
  v2fi = v2ii  -  uf * fabs( v2ft - v2it ) * td2 ;
  w1f  = w1i   -  r1 / i1 * uf * m1 * fabs( v1ft - v1it ) * td1 ;
  w2f  = w2i   +  r2 / i2 * uf * m2 * fabs( v2ft - v2it ) * td2 ;

  if ( ( v1fi + r1*w1f - v2pii ) * td1 < 0.0 ) overfric1 = 1; else overfric1 = 0;
  if ( ( v2fi - r2*w2f - v1pii ) * td2 < 0.0 ) overfric2 = 1; else overfric2 = 0;

  if( overfric1 || overfric2 )
  {
    if ( overfric1 ) 
      td1  = d1pii / ( uf * fabs( v1ft - v1it ) * (1.0 + m1*r1*r1/i1) ) ;
    if ( overfric2 )
      td2  = d2pii / ( uf * fabs( v2ft - v2it ) * (1.0 + m2*r2*r2/i2) ) ;
    if ( fabs(td1)<fabs(td2) ) td2=fabs(td1)*td2/fabs(td2) ;
    if ( fabs(td2)<fabs(td1) ) td1=fabs(td2)*td1/fabs(td1) ;
    v1fi = v1ii  -   uf * fabs( v1ft - v1it ) * td1 ;
    w1f  = w1i   -  r1 / i1 * uf * m1 * fabs( v1ft - v1it ) * td1 ;
    v2fi = v2ii  -   uf * fabs( v2ft - v2it ) * td2 ;
    w2f  = w2i   +  r2 / i2 * uf * m2 * fabs( v2ft - v2it ) * td2 ;
  }

  if ( uf!=0.0 && m1*v1fi*v1fi+i1*w1f*w1f+m2*v2fi*v2fi+i2*w2f*w2f>m1*v1ii*v1ii+i1*w1i*w1i+m2*v2ii*v2ii+i2*w2i*w2i )
  {
    float tildeke,a,b;
    b  = -2.0*m1*v1ii*(uf * fabs( v1ft - v1it ) * td1) -2.0*i1*w1i*(r1 / i1 * uf * m1 * fabs( v1ft - v1it ) * td1) ;
    b += -2.0*m2*v2ii*(uf * fabs( v2ft - v2it ) * td2) +2.0*i2*w2i*(r2 / i2 * uf * m2 * fabs( v2ft - v2it ) * td2) ;
    a  = m1*pow(uf * fabs( v1ft - v1it ) * td1,2) + i1*pow(r1 / i1 * uf * m1 * fabs( v1ft - v1it ) * td1,2) ;
    a += m2*pow(uf * fabs( v2ft - v2it ) * td2,2) + i2*pow(r2 / i2 * uf * m2 * fabs( v2ft - v2it ) * td2,2) ;
    tildeke = -b/a;
    if(tildeke<0.0)tildeke=0.0;
    v1fi = v1ii  -  uf * fabs( v1ft - v1it ) * td1 *tildeke ;
    w1f  = w1i   -  r1 / i1 * uf * m1 * fabs( v1ft - v1it ) * td1 *tildeke;
    v2fi = v2ii  -  uf * fabs( v2ft - v2it ) * td2 *tildeke;
    w2f  = w2i   +  r2 / i2 * uf * m2 * fabs( v2ft - v2it ) * td2 *tildeke;
  }

  // when player is manually turning, let the angular velocity change from this collision show 
  // by deactivating their turn command
  if( p1turnl && w1i!=0.0 && fabs(w1f/w1i-1.0)>=AVDE ) { p1turnl=3; } 
  if( p1turnr && w1i!=0.0 && fabs(w1f/w1i-1.0)>=AVDE ) { p1turnr=3; } 
  if( p2turnl && w2i!=0.0 && fabs(w2f/w2i-1.0)>=AVDE ) { p2turnl=3; }
  if( p2turnr && w2i!=0.0 && fabs(w2f/w2i-1.0)>=AVDE ) { p2turnr=3; }

  p1av = w1f;
  p2av = w2f;

  p1v = sqrt( v1ft*v1ft + v1fi*v1fi ) ;
  if ( v1fi!=0.0 || v1ft!=0.0 ) p1at = twopimod( atan2(v1fi,v1ft) + th1 );
  p2v = sqrt( v2ft*v2ft + v2fi*v2fi ) ;
  if ( v2fi!=0.0 || v2ft!=0.0 ) p2at = twopimod( atan2(v2fi,v2ft) + th2 );
}

void p1p2forwardafterbackout(float a, float dt)
{ //forward moves after collision bounce after back-out of pre-collision penetration
  moveplayer(&p1long,&p1lat,a*p1v,&p1at,&p1a,a*p1av,dt);
  moveplayer(&p2long,&p2lat,a*p2v,&p2at,&p2a,a*p2av,dt);
}

float p1p2backoutprecollpenet(float dist, float dt)
{ //backs out the pre-collision penetration of p1 and p2
  float th1,th2; //collision line bearings, theta
  float loni,lati,p12radii,ovrlp,fraction;
  float ac; float p1bat,p2bat,flt1;

  p12radii = (p1rad+p2rad)*M_PI/Y_RES;
  ovrlp = p12radii - dist ; 
  fraction = ( p1rad*M_PI/Y_RES - ovrlp*0.5 ) / dist ;
  intermediatepoint(p1long, p1lat, p2long, p2lat, fraction, &loni, &lati);
  th1 = angtofrom(loni, lati, p1long, p1lat) ;
  th2 = angtofrom(loni, lati, p2long, p2lat) ; 
  flt1 = p1v* cos ( p1at - th1 ) + p2v* cos ( p2at - th2 ) ;
  if(flt1<=0.0) return 0.0;
  ac = -ovrlp / flt1 ;
  p1bat=M_PI+p1at; p2bat=M_PI+p2at;
  moveplayer(&p1long,&p1lat,-ac*p1v,&p1bat,&p1a,ac*p1av,dt);
  moveplayer(&p2long,&p2lat,-ac*p2v,&p2bat,&p2a,ac*p2av,dt);
  p1at=-M_PI+p1bat; p2at=-M_PI+p2bat;
  if(ac>=-1.0) return -ac; else return 0.0;
}

void p1p2collisions(float dt)
{
  static int colliding=0;
  float dist;

  dist= metric( p1long,p1lat, p2long,p2lat );
  if (!colliding)
  {
    if ( dist < (p1rad+p2rad)*M_PI/Y_RES )
    {
      float backout; 
      colliding = 1;
      backout=p1p2backoutprecollpenet(dist,dt);
      //{ gotoxy(10,1); printf("%f",-backout); }
      //{ gotoxy(20,1); printf("%f",dist); }
      dist= metric( p1long,p1lat, p2long,p2lat ); //should be p1rad+p2rad after back-out
      //{ gotoxy(29,1); printf("%g",dist-(p1rad+p2rad)*M_PI/Y_RES); } 
      dist = (p1rad+p2rad)*M_PI/Y_RES;
      p1p2bounces(dist);
      p1p2forwardafterbackout(backout,dt);
    }
  }
  else
  {
    if ( dist < (p1rad+p2rad)*M_PI/Y_RES )
    {
      p1p2pushout4(dist);
      if ( metric( p1long,p1lat, p2long,p2lat ) >= (p1rad+p2rad)*M_PI/Y_RES )
        colliding=0;
    }
    else
      colliding=0;
  }
}

float schwarzchildr(float m)
{ // for CL not C, and not the universal speed limit
  return 2.0*GRAV*m*M_PI/Y_RES/CL/CL;
}

float blackabr2(float m)
{ // 6vinc player escape horizon from black abyss
  return 2.0*GRAV*m*M_PI/Y_RES/vinc/vinc/36.0;
}

void initblackab(void)
{
// for now I'll init it to lon/lat=0/0, but placing it away from current player positions might be good
  int i;
  blackab.lon = 0.0;
  blackab.lat = 0.0;
  blackab.v=0.0;
  blackab.at=0.0;
  blackab.am=0.0001;
  blackab.m= 10000;
  blackab.rad = schwarzchildr(blackab.m);
  blackab.rad2 = blackabr2(blackab.m);
  blackab.col= 252;//gridcol;
  for (i=0;i<BLACKABPIXGRABS;i++)
  {
    blackab.pglon[i]=1000.0;
    blackab.pglat[i]=0.0;
    blackab.pgv[i]=0.0;
    blackab.pgat[i]=0.0;
  }
}

void blackabgravity(float olon, float olat, float om, float *ov, float *oat, float dt, int real, int plrnotshot)
{
  float fg, r, th, flt1, flt2;
  r = metric(olon,olat,blackab.lon,blackab.lat);
  if (r>=blackab.rad)
  {
    float blackabm=blackab.m*M_PI/Y_RES;
    om *= M_PI/Y_RES;
    if( blackab.exists==1 || metric!=spheremetric )
      fg = GRAV*blackabm*om/r/r ;
    else
    {
      vector d;
      float ar;
      d = cartd(olon,olat,blackab.lon,blackab.lat);
      ar = vectdot(d,d);
      d = vectcross(d,cartr(olon,olat));
      //d = vectdiff( d, vecttimes( vectdot(d,cartr(olon,olat)), d ) ) ; //same magnitude
      fg = GRAV*blackabm*om/ar*sqrt(vectdot(d,d)) ;
    }
    th = angtofrom(blackab.lon,blackab.lat,olon,olat);
    if ( metric==spheremetric && plrnotshot)
      flt1 = *ov * cos(*oat-th) + contactfrictionorbitalmomentarm(om)*fg/playerorbitmomentsphere(om)*dt;
    else
      flt1 = *ov * cos(*oat-th) + fg/om*dt;
    flt2 = *ov * sin(*oat-th) ;
    *ov = sqrt( flt1*flt1 + flt2*flt2 ) ;
    if ( flt1!=0.0 || flt2!=0.0 ) *oat = twopimod( atan2(flt2,flt1) + th );
    if (real)
    {
      th = angtofrom(olon,olat,blackab.lon,blackab.lat);
      flt1 = blackab.v * cos(blackab.at-th) + fg/blackabm*dt;
      flt2 = blackab.v * sin(blackab.at-th) ;
      blackab.v = sqrt( flt1*flt1 + flt2*flt2 ) ;
      if ( flt1!=0.0 || flt2!=0.0 ) blackab.at = twopimod( atan2(flt2,flt1) + th );
    }
  }
}

void blackabgravitar(float dt)
{
  int i;
  if(p1exists) blackabgravity(p1long,p1lat,p1rad,&p1v,&p1at,dt,1,1);
  if(p2exists) blackabgravity(p2long,p2lat,p2rad,&p2v,&p2at,dt,1,1);
  for (i=0;i<MAX_SHOTS;i++)
  {
    if ( (shots[i].owner)!=-1)
      blackabgravity(shots[i].lon,shots[i].lat,0.5,&(shots[i].v),&(shots[i].bear),dt,1,0);
  }
}

void baabsorbmomentum(float plon, float plat, float pm, float pv, float pat, int plrnotshot)
{
  float totm;
  if(metric==spheremetric)
  {
    vector mergv; //velocity vector of player after merged with shot
    totm = pm + blackab.m ;
    if(plrnotshot) pm=playerorbitmomentsphere(pm*M_PI/Y_RES)/M_PI*Y_RES; 
    mergv=wcrossr(vecttimes(1.0/totm,vectsum( vecttimes(pm,rcrossv(plon,plat,pv,pat)),
           vecttimes(blackab.m,rcrossv(blackab.lon,blackab.lat,blackab.v,blackab.at)) ) ),blackab.lon,blackab.lat);
    blackab.m=totm;
    bearchang1(blackab.lon,blackab.lat,mergv.x,mergv.y,mergv.z,&(blackab.at),&totm); // at = atan( (mergv dot lat hat) / (mergv dot lon hat) ), totm used as dummy
    blackab.v=sqrt(vectdot(mergv,mergv));
  }
  else 
  {
    float vl,vf,vsl,vsf;
    totm = pm + blackab.m ;
    vl = pv * cos(pat) ;
    vf = pv * sin(pat) ;
    vsl = blackab.v * cos( blackab.at ) ;
    vsf = blackab.v * sin( blackab.at ) ;
    vl = ( pm * vl + blackab.m * vsl ) / totm ; 
    vf = ( pm * vf + blackab.m * vsf ) / totm ;
    blackab.m=totm;
    blackab.v = sqrt (vl*vl + vf*vf) ;
    if ( vl != 0.0 || vf != 0.0 ) blackab.at = atan2(vf,vl) ;
  }
}

void blackabcollisions(void)
{  // haven't added mass+momentum absorption yet
  int i;
  if(p1exists)
    if(metric(p1long,p1lat,blackab.lon,blackab.lat)<blackab.rad)
    {
      baabsorbmomentum(p1long,p1lat,p1rad,p1v,p1at,1);
      blackab.am+=playermoment(p1rad*M_PI/Y_RES)*p1av;
      deexist_player1();
    }
  if(p2exists)
    if(metric(p2long,p2lat,blackab.lon,blackab.lat)<blackab.rad)
    {
      baabsorbmomentum(p2long,p2lat,p2rad,p2v,p2at,1);
      blackab.am+=playermoment(p2rad*M_PI/Y_RES)*p2av;
      deexist_player2();
    }
  for (i=0;i<MAX_SHOTS;i++)
  {
    if ( (shots[i].owner)!=-1)
      if(metric(shots[i].lon,shots[i].lat,blackab.lon,blackab.lat)<blackab.rad)
      {
        baabsorbmomentum(shots[i].lon,shots[i].lat,shots[i].rad,shots[i].v,shots[i].bear,0);
        shots[i].owner=-1;
        //should shot explode instead?
        //shots[i].exploding=1;
      }
  }
}

void drawblackab(float dt)
{
  int x,y,c,bc;
  float th, rp; //parameters swept
  float lonp, latp;  //lon,lat parameters
  int j;
  int k;
  float lon=blackab.lon;
  float lat=blackab.lat;
  float r1=blackab.rad;
  float r2=blackab.rad2;
  float r2r1=r2-r1;

  projec(lon,lat,&x,&y);
  if(backproj && blackab.col>BACKPROJDIFF) bc=blackab.col-BACKPROJDIFF; else bc=blackab.col;
  {
    //disappear pixels within r1
    {
      th = rand()*M_PIT2/RAND_MAX;
      rp = rand()*r1/RAND_MAX;
      surfacerun(lon, lat, th, rp, &lonp, &latp);
      projec(lonp,latp,&x,&y);
      PUTPIXS(x,y,0);
    }
    //grab pixels between r1 and r2 for gravitation
    k=0;
    while(k<BLACKABPIXGRABS)
    {
      if(blackab.pglon[k]==1000.0) break;
      k++;
    }
    if(k<BLACKABPIXGRABS)
    {
      c=0; j=0;
      while(c==0 && j<25)
      {
        th = rand()*M_PIT2/RAND_MAX;
        rp = rand()*r2r1/RAND_MAX + r1;
        surfacerun(lon, lat, th, rp, &lonp, &latp);
        projec(lonp,latp,&x,&y);
        c=GETPIXS(x,y); 
        j++;  
      }
      if (c!=0) 
      {
        blackab.pglon[k]=lonp;
        blackab.pglat[k]=latp;
        blackab.pgv[k]=0.0;
        blackab.pgat[k]=0.0;
        blackab.pgc[k]=c;
      }
    }
    //gravitate grabbed pixels
    for(k=0;k<BLACKABPIXGRABS;k++)
    {
      lonp=blackab.pglon[k];
      if(lonp!=1000.0)
      {
        float flt;
        latp=blackab.pglat[k];
        projec(lonp,latp,&x,&y);
        PUTPIXS(x,y,0); 
        blackabgravity(lonp,latp,0.3,&(blackab.pgv[k]),&(blackab.pgat[k]),dt,0,0);
        moveplayer(&lonp,&latp,blackab.pgv[k],&(blackab.pgat[k]),&th,th,dt);
        projec(lonp,latp,&x,&y);
        flt=metric(lonp,latp,lon,lat);
        if(flt>blackab.rad2)
        {
          blackab.pglon[k]=1000.0;
        }
        else
        {
          PUTPIXS(x,y,blackab.pgc[k]); 
          if(flt<blackab.rad)
          {
            blackab.pglon[k]=1000.0;
          }
          else
          {
            blackab.pglon[k]=lonp;
            blackab.pglat[k]=latp;
          }
        }
      }
    }
    //generate grabbed pixels within r2
    if ( rand()%100>45 )
    {
      k=0;
      while(k<BLACKABPIXGRABS)
      {
        if(blackab.pglon[k]==1000.0) break;
        k++;
      }
      if(k<BLACKABPIXGRABS)
      {
        th = rand()*M_PIT2/RAND_MAX;
        rp = (rand()*0.1/RAND_MAX+0.88)*r2;
        surfacerun(lon, lat, th, rp, &lonp, &latp);
        projec(lonp,latp,&x,&y);
        PUTPIXS(x,y,bc); 
        blackab.pglon[k]=lonp;
        blackab.pglat[k]=latp;
        blackab.pgv[k]=0.0;
        blackab.pgat[k]=0.0;
        blackab.pgc[k]=bc;
      }
    }
  }
}

void setprojec(int projecti)
{
        switch(projecti)
        {
          case 0: projec=sinproj; /*invprojec=invsinproj;*/ break;
          case 1: projec=winktrproj; break;
          case 2: projec=rectproj; break;
          case 3: projec=merctrproj; /*invprojec=invmerctrproj;*/ break;
          case 4: projec=twohemiproj; break;
          case 5: projec=twohemiproj2; break;
          case 6: projec=sphereproj; break;
          case 7: projec=perspsphereproj; break;
          case 8: projec=azimuequiproj; break;
          case 9: projec=albersconicproj; break;
        }
}

void morehelp( int nogrid, int logrid, int kollisions, int shotplrkollisions, int rectilinearmode, int physdisp, int movestiming )
{
//  clrscr();
  gotoxy(1,1);
  textbackground( LIGHTGRAY );
  cprintf(" ");
  textbackground( BLACK ); cprintf("         ");
  textcolor( BLACK ) ;
  textbackground( CYAN );
  cprintf("Commands Help and Status Screen 2\n");
  textcolor( LIGHTGRAY ) ;
  textbackground( BLACK );
  gotoxy(1,wherey());
  printf("\n");
  printf("[key]: effect or status\n");

  printf("9: player 1 fire. detonate oldest shot if 3 shots are already out.\n");
  printf("8: detonate player 1 shot nearest to player 1.\n");
  printf("e: player 2 fire. detonate oldest shot if 3 shots are already out.\n");
  printf("w: detonate player 2 shot nearest to player 2.\n");
  printf("k or K: toggle player-player and player-shot collisions, %d %d.\n",kollisions,shotplrkollisions);
  printf("; or :: animated transition from one projection to another.\n");
  printf("{: rectilinear mode, ");
  printf("rectangle space, not a sphere, %d.  \n",rectilinearmode);
  printf("}: mode-independent selection of spherical or rectilinear space: ");
  if(metric==spheremetric)       printf("spherical.   \n");
  if(metric==squaremetric)       printf("rectilinear. \n");
  if(metric==walledsquaremetric) printf("walled rect. \n");
  printf("0: black hole (with two interpretations of gravity in spherical space), %d.\n",blackab.exists);
  printf("6: no grid lines except around the border of the projection, %d.\n",nogrid);
  if(logrid<=300)
  printf("g or G: modify grid color. grid color is %3d.  ^: grid refresh delay, %3d/18 s.\n",gridcol,logrid);
  else
  printf("g or G: modify grid color. grid color is %3d.  ^: grid refresh delay, forever. \n",gridcol,logrid);
  printf("v or V: player 2 colors, %3d %3d.  ",p2colc,p2colf);
  printf("b or B: player 1 colors, %3d %3d.\n",p1colc,p1colf);
  printf("h or H: restore or save player coordinates.\n");
  printf("]: try to adjust magnitude of movement steps in balance with program speed, %d.\n",movestiming);
  printf("[: kinetic energy and momentum read-out at bottom of screen, %d.\n",physdisp);
  printf("\\: frames-per-second read-out.\n");
  printf("? or /: enter or exit this screen.\n");
  textbackground( LIGHTGRAY );
  gotoxy(79,24);
  cprintf(" ");
  textcolor( LIGHTGRAY ) ;
  textbackground( BLACK );
  gotoxy(1,24);
  printf("  ((()))  ");
  gotoxy(79,24);
}

void helpandstat(int projecti, int delnum, int objdrawselect, int autoclear, int friction, int p2pfriction, 
  int shotkollangular, int erasetrails, int eraseshottrails, int mixingprojs, int rectilinearmode, int physdisp, int wallspin, int betterfriction)
// commands help, and status
{
  int i;
  //clrscr();
  gotoxy(1,1);
  textbackground( LIGHTGRAY );
  cprintf(" ");
  gotoxy(11,1);
  textcolor( BLACK ) ;
  textbackground( CYAN );
  cprintf("Commands Help and Status Screen\n");
  textcolor( WHITE ) ;
  textbackground( LIGHTGRAY );
  //gotoxy(11,2); cprintf("                               ");
  textcolor( LIGHTGRAY ) ;
  textbackground( BLACK );
  gotoxy(1,wherey());
  printf("\n");
  printf("[key]: effect or status\n");
  printf("z x c a s d: player 2 turn left, forward, turn right, left, back, right.\n");
  printf("1 2 3 4 5 6 (on numpad): player 1 movement keys, like player 2.\n");
  printf("p or P: select type of projection: ");
  if(rectilinearmode)
  printf("n/a (currently in rectilinear mode).\n"); 
  else
      switch(projecti){
          case 0:
  if(mixingprojs) 
  printf("(mixing to) sinusoidal projection.          \n");
  else
  printf("sinusoidal projection.                      \n"); break;
          case 1:
  if(mixingprojs) 
  printf("(mixing to) Winkel Tripel.                  \n");
  else
  printf("Winkel Tripel.                              \n"); break;
          case 2:
  if(mixingprojs) 
  printf("(mixing to) rectangular.                    \n"); 
  else
  printf("rectangular.                                \n"); break;
          case 3:
  if(mixingprojs) 
  printf("(mixing to) Mercator.                       \n"); 
  else
  printf("Mercator.                                   \n"); break;
          case 4:
  if(mixingprojs) 
  printf("(mixing to) two hemispheres, flat.          \n"); 
  else
  printf("two hemispheres, flat.                      \n"); break;
          case 5:
  if(mixingprojs) 
  printf("(mixing to) two hemispheres, one from under.\n"); 
  else
  printf("two hemispheres, one from under.            \n"); break;
          case 6:
  if(mixingprojs) 
  printf("(mixing to) flat.                           \n"); 
  else
  printf("flat.                                       \n"); break;
          case 7:
  if(mixingprojs) 
  printf("(mixing to) perspective.                    \n"); 
  else
  printf("perspective.                                \n"); break;
          case 8:
  if(mixingprojs) 
  printf("(mixing to) azimuthal equidistant.          \n"); 
  else
  printf("azimuthal equidistant.                      \n"); break;
          case 9:
  if(mixingprojs) 
  printf("(mixing to) Alber's Equal-Area Conic.       \n"); 
  else
  printf("Alber's Equal-Area Conic.                   \n"); break;
  }  
  printf("t or T: trails or erasure of previous player or shot positions, %d %d.\n",erasetrails,eraseshottrails);
  printf("5 or %%: decrease or increase main loop delay, %2dms.  \n",delnum);
  printf("o or O: draw player shapes projected like the space, or not: %d.\n",objdrawselect);
  printf("` or -: clear the screen.      ~: automatic screen clearing, %d.\n",autoclear);
  printf("7: reset player 2.             8: reset player 1.\n");
  printf("=: reset board (players to starting coordinates and a few other resets).\n");
  printf("': curvilinear trajectory math method, %d, ",trajmath);
  switch(trajmath%TRAJMATHS) {
    case 0: printf("centripetal step.        \n"); break;
    case 1: printf("parametric great circle. \n"); break;
    case 2: printf("solved great circle run. \n"); break;
  }
  printf("l or L: spin mechanics in player-player and player-shot collisions, %d %d.\n",p2pfriction,shotkollangular);
  printf("u: playfield friction on saucer travel, %d.  ",friction);
  printf("I: wall-player friction spin, %d.\n",wallspin);
  printf("(spacebar): pause.  U: enable more realistic playfield surface friction, %d.\n",betterfriction);
  printf("(escape): quit program or exit this screen.\n");
  printf("?: enter or exit this screen.  ");
  printf("/: show help on more commands.\n");
  printf("\n");
  printf("lon,lat %6.3f, %6.3f  look angle %4.2f  travel angle %4.2f  vel %6.4f   \n",p1long,p1lat,p1a,p1at,p1v*100);
  printf("lon,lat %6.3f, %6.3f  look angle %4.2f  travel angle %4.2f  vel %6.4f   \n",p2long,p2lat,p2a,p2at,p2v*100);
  printf("\n");
  printf("          ");
  textcolor( BLACK ) ;
  textbackground( CYAN );
  if(!physdisp) cprintf("Spherical Spacewar, 6/15/06 to 8/15/06        --Allen Jordan\n");
  textbackground( LIGHTGRAY );
  gotoxy(79,24);
  cprintf(" ");
  textcolor( LIGHTGRAY ) ;
  textbackground( BLACK );
  gotoxy(1,24);
  printf("  ((()))  ");
  gotoxy(79,24);
}

void walledsquaresurface()
{
  metric = walledsquaremetric ;
  surfacerun = walledsquarerun ;
  intermediatepoint = walledrectintermediatepoint; 
  angtofrom = walledsquarecoursetofrom; 
  moveplayer = walledsquaremoveplayer ;
  playermoment = playermomentflat ;
  absorbknockturn = flatabsorbknockturn;
  p1p2bounces = p1p2bounce_flat;
  frictionintegral = flatfrictionint;
}

void squaresurface()
{
  metric = squaremetric ;
  surfacerun = squarerun ;
  intermediatepoint = rectintermediatepoint ;
  angtofrom = squarecoursetofrom ;  
  moveplayer = squaremoveplayer ;
  playermoment = playermomentflat ;
  absorbknockturn = flatabsorbknockturn;
  p1p2bounces = p1p2bounce_flat;
  frictionintegral = flatfrictionint;
}

void spheresurface()
{
  metric = spheremetric ;
  surfacerun = sphererun ;
  intermediatepoint = sphereintermediatepoint ;
  angtofrom = coursetofrom ;  
  moveplayer = spheremoveplayer ;
  playermoment = playermomentsphere ;
  absorbknockturn = sphereabsorbknockturn;
  p1p2bounces = p1p2bounce_sphere;
  frictionintegral = spherefrictionint;
}

int main(void)
{
  int quit=0;
  int projecti;
  int ch, i;
  int fire1=0,fire2=0;
  int detn1=0,detn2=0; //detonate nearest precaution
  float flt1,flt2,flt3; //temp floats
  int objdrawchngflag=0, objdrawselect;
  int erasetrails=3;
  int eraseshottrails=1;
  int kollisions=1;
  int p2pfriction=1;
  int shotplrkollisions=1;
  int shotkollangular=1;
  int friction=1;
  int betterfriction=0;
  int wallspin=1;
  int nogrid=1; // border only grid display
  int logrid=105; // clock ticks grid refresh delay
  int delnum=1;
  int delnummem=1;
  int intextscr=0;
  int pause=0;
  int autoclear=0;
  int startmixingprojs=0, mixingprojs=0, pausemixingprojs=0 ; // flags for while performing animated transition between two projections
  unsigned long int loopctr=0;// loop counts so far // at 300 loops per second, would take almost half a year to flip
  unsigned long int prevtickloopct=0;
  int prevloopcts=16; // how many loops in the last clock tick
  int smoothedloopcts[16]={16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16}; // smoothed previous loop counts; earlier method (15* + new)/16 seemed odd
  int smoothedlci=1; //index into smoothedloopcts[]
  clock_t start, tick, tock, gtick, keytick, dptick;
  int kwait[16]; //for use with delaying repeat of player control keys
  int rectilinearmode=0,startrectilinearmode=0,exitrectilinearmode=0,changespace=0;
  int physdisp=0;
  int blankphysdisp=0;
  int fpsreadout=0;
  int movestiming=0;  // option to scale moves speeds by LOOPSPTICK/prevloopcts
  int abortplayer1=0, abortplayer2=0;
  blackab.exists=0;

  for ( ch=0; ch<16; ch++ ) kwait[ch]=0;
  start = clock();
  dptick = keytick = gtick = clock();
  projecti=2;
  projec=rectproj;
  //invprojec=invsinproj;
  objdrawselect=1;
  drawobj = drawplayernsr;
  playerwallbounces = playerwallbouncel; //for rectilinearmode==2
  muzzlespeed=vinc*2.8;
  shotexplrad=10;

  spheresurface();

  keyp1tl=keyp1f=keyp1tr=keyp1l=keyp1b=keyp1r=keyp1fire=keyp1det=0;
  keyp2tl=keyp2f=keyp2tr=keyp2l=keyp2b=keyp2r=keyp2fire=keyp2det=0;
  keyenvintcom=0;

  canvas = SCREEN;
  //scrnbuff = malloc((unsigned)64000);
  //if (scrnbuff==NULL) { printf("null malloc");pauser(&i);exit(0);}
  vga_mode(0x13);
  vgapalset();
  //clear_screen_buff();
  clear_screen();
  //update_screen();

  init_keyboard();

  init_player1();
  init_player2();
  //p1rad=p2rad=30; //93;
  p1rad=30;p2rad=155; //93;
  p1colc=191;
  saveplayers();
  init_player1();
  init_player2();
  p1shotsfired = 0;
  p2shotsfired = 0;
  if(nogrid) drawgridbo(1); else drawgrid(1);
  initshots();

  tick = clock();
  while (!quit)
  {
    delay(delnum);

    loopctr++;
    if ( ( tock=clock() ) > tick ) // how long does it take for clock to flip a u long integer? 7 years.
    {
      prevloopcts = loopctr - prevtickloopct;  // (int)(loopctr - prevtickloopct);
      ch=0; for(i=1;i<16;i++) ch+=smoothedloopcts[i];
      smoothedloopcts[0] = (ch+prevloopcts)>>4; smoothedloopcts[smoothedlci]=prevloopcts;
      smoothedlci = ( smoothedlci + 1 )%15 + 1;
      prevtickloopct = loopctr;
      tick = tock;
    } //system clock() ticks ~18.2 times a second

    if(!intextscr) if ((tock>=gtick+logrid&&logrid<=300)||gtick==0) { gtick = clock(); if(nogrid || rectilinearmode) drawgridbo(1); else drawgrid(1); }
    if (pause) { if(intextscr)textpauser(&pause,&keytick); else pauser(&pause); }

    if ( keyenvintcom || keyboardproc() )
    {
      ch=keyenvintcom;
      keyenvintcom=0;
      keytick = clock();
    switch( ch )
    {
      case 0x1b:
      { if(!intextscr) quit=1;
        else
        {
          intextscr = 0;
          vga_mode(0x13);
          vgapalset();
        }
      }
      break;
      case 'P':
      case 'p': if(!mixingprojs && !rectilinearmode) 
      {
        if (ch=='p')        projecti++;
        if (ch=='P')        projecti--;
        projecti = imod(projecti,PROJECTIONS);
        setprojec(projecti);
        if(!intextscr) { clear_screen(); gtick=0; }
        flushshotstreaks();
        objdrawchngflag=1;
      }
      break;
      case ';':  //mixproj
      if (!rectilinearmode)
      {
        if(!mixingprojs) startmixingprojs = 1 ; else pausemixingprojs=!pausemixingprojs ;
      }
      break;
      case ':':  //mixproj   
      if (!rectilinearmode)
      {
        if(!mixingprojs) startmixingprojs = 2 ; else pausemixingprojs=!pausemixingprojs ;
      }
      break;
      case '{':
        if(!rectilinearmode) 
        {
          if(metric==walledsquaremetric)
            startrectilinearmode=2; 
          else 
            startrectilinearmode=1; 
          mixingprojs=pausemixingprojs=0; 
        }
        else 
        if(rectilinearmode==1) 
         { rectilinearmode=2; //with walls instead of wrapping
          changespace=3; }
        else 
          exitrectilinearmode=1;
      break;
      case '}': if(metric==spheremetric) { changespace=2; if(rectilinearmode==2) rectilinearmode=1; }
                else if(metric==squaremetric) { changespace=3; if(rectilinearmode==1) rectilinearmode=2; }
                else if(metric==walledsquaremetric) { changespace=1; if(rectilinearmode==2) rectilinearmode=1; }
        // to try rectilinear space with 'sphere' projections
      break;
      case '0': if(!blackab.exists)initblackab(); blackab.exists=( blackab.exists + 1)%3; 
      break;
      case 't': erasetrails=(erasetrails+1)%ERASES; 
      break;
      case 'T': eraseshottrails=!eraseshottrails;
      break;
      case 'k': kollisions=!kollisions;
      break;
      case 'K': shotplrkollisions=!shotplrkollisions;
      break;
      case 'u': friction=(friction+1)%4;
      break;
      case 'U': betterfriction=!betterfriction;
      break;
      case 'I': wallspin=(wallspin+1)%2;
        switch(wallspin%2)
        {
          case 0:   playerwallbounces = playerwallbounce; break;
          case 1:   playerwallbounces = playerwallbouncel; break;
        }
      break;
      case '\'': { trajmath=(trajmath+1)%TRAJMATHS;
                    }
      break;
      case ' ':
      {
        pause=1;
      }
      break;
      case '?':
      {
        if(!intextscr)
        {
          vga_mode(0x03);
          intextscr=1;
        }
        else
        if(intextscr==1)
        {
          intextscr = 0;
          vga_mode(0x13);
          vgapalset();
          gtick=0;
        }
        else
        if(intextscr==2) { intextscr=1; clrscr(); }
      }
      break;
      case '/':
      {
        if(!intextscr)
        {
          vga_mode(0x03);
          intextscr=1;
        }
        else
        if(intextscr==1)
        {
          intextscr=2; clrscr();
        }
        else
        if(intextscr==2) { intextscr=1; clrscr(); }
      }
      break;
      case '%': delnum++;
      break;
      case '5': {if (delnum!=0) {delnummem=delnum; delnum=0;}
                    else { delnum=delnummem; }  }
      break;
      case '\\': fpsreadout=!fpsreadout;
      break;
      case '|': curioswitch2=!curioswitch2;
      break;
      case '[': physdisp=!physdisp; 
        if (physdisp)
        {
          if (intextscr==1) { gotoxy(11,23); printf("                                                            \n"); } 
        }
        else 
        {
          blankphysdisp=1; 
          if (!intextscr) gtick=0; 
        }
      break;
      case ']': movestiming=!movestiming;
      break;
      case 'l':
      {
        p2pfriction=(p2pfriction+1)%2;
        switch(p2pfriction%2)
        {
          case 0:   uf_p2p=0.0; break;
          case 1:   uf_p2p=0.1; break;
        }
      }
      break;
      case 'L': shotkollangular=!shotkollangular;
      break;
      case 'O': { objdrawchngflag=1; objdrawselect=(objdrawselect+OBJDRAWS-1)%OBJDRAWS;}
      break;
      case 'o': { objdrawchngflag=1; objdrawselect=(objdrawselect+1)%OBJDRAWS;}
      break;
      case 'g': gridcol=rand()%256; gtick=0; 
      break;
      case 'G': gridcol=(gridcol+1)%256; gtick=0; 
      break;
      case '^': logrid=(logrid+100)%400;
      break;
      case '6': nogrid=!nogrid; 
        if (!intextscr)
        {
          if(nogrid) drawgrid(0);
          gtick=0;
        }
      break;
      case 'v': p2colc=(p2colc+16)%256;
      break;
      case 'V': p2colf=(p2colf+16)%256;
      break;
      case 'b': p1colc=(p1colc+16)%256;
      break;
      case 'B': p1colf=(p1colf+16)%256;
      break;
      case 'h': restoreplayers();
      break;
      case 'H': saveplayers();
      break;
      case '`':
      case '-': 
        if (intextscr) clrscr();
        else { clear_screen(); gtick=0; }
      break;
      case '~': autoclear=(autoclear+1)%8;
      break;
      case '8': init_player1();
      break;
      case '7': init_player2();
      break;
      case '*': abortplayer1=1;
      break;
      case '&': abortplayer2=1;
      break;
      case '=':
      {
        init_player1();
        init_player2();
        detall(0); detall(1); //initshots();
        if (intextscr) clrscr();
        else { clear_screen(); gtick=0; }
        if( !intextscr ) vgapalset();
        fpsreadout=0; 
        curioswitch2=0; 
        if( !rectilinearmode && metric!=spheremetric ) changespace=1;
        if( rectilinearmode && metric==spheremetric ) changespace=rectilinearmode+1;
        if( logrid>300 ) logrid=105;
      }
      break;
    } // end switch( ch )
    }
    for ( i=0; i<16; i++ )
      if (kwait[i]>=0) kwait[i]--;
    //#define TURNWAIT 0
    #define MOVEWAIT 1
    #define FIREWAIT 2
    if(p1exists)
    {
      if (keyp1tl && !p1turnl)
      {
        p1turnl=1;
        if(p1turnr) { p1turnl=2; p1turnr=1; }
        //kwait[0]=prevloopcts>>TURNWAIT;
      }
      if (keyp1tr && !p1turnr)
      {
        p1turnr=1;
        if(p1turnl) { p1turnr=2; p1turnl=1; }
        //kwait[2]=prevloopcts>>TURNWAIT;
      }
      if (keyp1f && kwait[1]<0)
      {
        flt3=(p1v*cos(p1at)+vinc*cos(p1a));
        flt2=(p1v*sin(p1at)+vinc*sin(p1a));
        if(flt2!=0.0||flt3!=0.0)p1at=atan2(flt2,flt3);
        p1v=sqrt(flt3*flt3+flt2*flt2);
        kwait[1]=prevloopcts<<MOVEWAIT; //keyp1f=0;
      }
      if (keyp1b && kwait[4]<0)
      {
        flt3=(p1v*cos(p1at)-vinc*cos(p1a));
        flt2=(p1v*sin(p1at)-vinc*sin(p1a));
        if(flt2!=0.0||flt3!=0.0)p1at=atan2(flt2,flt3);
        p1v=sqrt(flt3*flt3+flt2*flt2);
        kwait[4]=prevloopcts<<MOVEWAIT; //keyp1b=0;
      }
      if (keyp1l && kwait[3]<0)
      {
        flt3=(p1v*cos(p1at)-vinc*sin(p1a));
        flt2=(p1v*sin(p1at)+vinc*cos(p1a));
        if(flt2!=0.0||flt3!=0.0)p1at=atan2(flt2,flt3);
        p1v=sqrt(flt3*flt3+flt2*flt2);
        kwait[3]=prevloopcts<<MOVEWAIT; //keyp1l=0;
      }
      if (keyp1r && kwait[5]<0)
      {
        flt3=(p1v*cos(p1at)+vinc*sin(p1a));
        flt2=(p1v*sin(p1at)-vinc*cos(p1a));
        if(flt2!=0.0||flt3!=0.0)p1at=atan2(flt2,flt3);
        p1v=sqrt(flt3*flt3+flt2*flt2);
        kwait[5]=prevloopcts<<MOVEWAIT; //keyp1r=0;
      }
      if (keyp1fire && kwait[6]<0)
      {
        fire1=1;
        kwait[6]=prevloopcts<<FIREWAIT; //keyp1fire=0;
      }
      if (keyp1det && kwait[7]<0)
      {
        detn1=1;
        kwait[7]=prevloopcts*2.67;
      }
    }
    if(p2exists)
    {
      if (keyp2tl && !p2turnl)
      {
        p2turnl=1;
        if(p2turnr) { p2turnl=2; p2turnr=1; }
        //kwait[8]=prevloopcts>>TURNWAIT;
      }
      if (keyp2tr && !p2turnr)
      {
        p2turnr=1;
        if(p2turnl) { p2turnr=2; p2turnl=1; }
        //kwait[10]=prevloopcts>>TURNWAIT;
      }
      if (keyp2f && kwait[9]<0)
      {
        flt3=(p2v*cos(p2at)+vinc*cos(p2a));
        flt2=(p2v*sin(p2at)+vinc*sin(p2a));
        if(flt2!=0.0||flt3!=0.0)p2at=atan2(flt2,flt3);
        p2v=sqrt(flt3*flt3+flt2*flt2);
        kwait[9]=prevloopcts<<MOVEWAIT; //keyp2f=0;
      }
      if (keyp2b && kwait[12]<0)
      {
        flt3=(p2v*cos(p2at)-vinc*cos(p2a));
        flt2=(p2v*sin(p2at)-vinc*sin(p2a));
        if(flt2!=0.0||flt3!=0.0)p2at=atan2(flt2,flt3);
        p2v=sqrt(flt3*flt3+flt2*flt2);
        kwait[12]=prevloopcts<<MOVEWAIT; //keyp2b=0;
      }
      if (keyp2l && kwait[11]<0)
      {
        flt3=(p2v*cos(p2at)-vinc*sin(p2a));
        flt2=(p2v*sin(p2at)+vinc*cos(p2a));
        if(flt2!=0.0||flt3!=0.0)p2at=atan2(flt2,flt3);
        p2v=sqrt(flt3*flt3+flt2*flt2);
        kwait[11]=prevloopcts<<MOVEWAIT; //keyp2l=0;
      }
      if (keyp2r && kwait[13]<0)
      {
        flt3=(p2v*cos(p2at)+vinc*sin(p2a));
        flt2=(p2v*sin(p2at)-vinc*cos(p2a));
        if(flt2!=0.0||flt3!=0.0)p2at=atan2(flt2,flt3);
        p2v=sqrt(flt3*flt3+flt2*flt2);
        kwait[13]=prevloopcts<<MOVEWAIT; //keyp2r=0;
      }
      if (keyp2fire && kwait[14]<0)
      {
        fire2=1;
        kwait[14]=prevloopcts<<FIREWAIT; //keyp2fire=0;
      }
      if (keyp2det && kwait[15]<0)
      {
        detn2=1;
        kwait[15]=prevloopcts*2.67;
      }
    }
      //if there's a turn command and its key goes up and the other turn command is in the higher priority state, demote that state
      if (p1turnl && !keyp1tl && p1turnr==2) p1turnr=1;
      if (p1turnr && !keyp1tr && p1turnl==2) p1turnl=1;
      if (p2turnl && !keyp2tl && p2turnr==2) p2turnr=1; 
      if (p2turnr && !keyp2tr && p2turnl==2) p2turnl=1;
      //if there's a turn command and its not in post-collision hold and its key goes up, zero the turn command and if neither turn key is down, zero the spin and quantize the look angle
      if (p1turnl && p1turnl<3 && !keyp1tl) { p1turnl=0; if(!keyp1tr){p1av=0.0; p1a=nearfmodmult(p1a,ainc);} } 
      if (p1turnr && p1turnr<3 && !keyp1tr) { p1turnr=0; if(!keyp1tl){p1av=0.0; p1a=nearfmodmult(p1a,ainc);} } 
      if (p2turnl && p2turnl<3 && !keyp2tl) { p2turnl=0; if(!keyp2tr){p2av=0.0; p2a=nearfmodmult(p2a,ainc);} } 
      if (p2turnr && p2turnr<3 && !keyp2tr) { p2turnr=0; if(!keyp2tl){p2av=0.0; p2a=nearfmodmult(p2a,ainc);} } 
      //if a turn command is in post-collision hold and its key goes up, zero the turn command
      if (p1turnl==3 && !keyp1tl) p1turnl=0;
      if (p1turnr==3 && !keyp1tr) p1turnr=0;
      if (p2turnl==3 && !keyp2tl) p2turnl=0;
      if (p2turnr==3 && !keyp2tr) p2turnr=0;

    if(erasetrails&&!intextscr)            //erase previous player draws
    {
      if(erasetrails==1)
      {
        if(p1exists&&(p1v!=0.0 || mixingprojs)) drawobj(p1long, p1lat, p1a, p1rad, 0, 0);
        if(p2exists&&(p2v!=0.0 || mixingprojs)) drawobj(p2long, p2lat, p2a, p2rad, 0, 0);
      } else
      if(erasetrails==2)
      {
        if(p1exists&&(p1v!=0.0 || mixingprojs)) eraseplayernsrte(p1long, p1lat, p1a, p1rad, p1v, p1at);
        if(p2exists&&(p2v!=0.0 || mixingprojs)) eraseplayernsrte(p2long, p2lat, p2a, p2rad, p2v, p2at);
      } else
      if(erasetrails==3)
      {
        if(p1exists&&(p1v!=0.0 || mixingprojs)) dpnsrerasering(p1long, p1lat, p1a, p1rad, 0, 0);
        if(p2exists&&(p2v!=0.0 || mixingprojs)) dpnsrerasering(p2long, p2lat, p2a, p2rad, 0, 0);
      } 
    }
    if(eraseshottrails&&!intextscr)        eraseshots(); //erase previous draws

    if (fire1) {shoot(0); fire1=0; }
    if (fire2) {shoot(1); fire2=0; }
    if (detn1) {detnearest(0); detn1=0; }
    if (detn2) {detnearest(1); detn2=0; }
    if (abortplayer1) { if(p1exists){drawobj(p1long, p1lat, p1a, p1rad, 0, 0);deexist_player1();} abortplayer1=0; }
    if (abortplayer2) { if(p2exists){drawobj(p2long, p2lat, p2a, p2rad, 0, 0);deexist_player2();} abortplayer2=0; }

    if( movestiming ) 
    {
      if (p1turnl && p1turnl<3 && p1turnl>p1turnr) { p1av= ainc*1.5/LOOPSPTICK; } // av and p1turnl are zeroed on key up
      if (p1turnr && p1turnr<3 && p1turnr>p1turnl) { p1av=-ainc*1.5/LOOPSPTICK; } //  elsewhere, in the input section of the loop
      if (p2turnl && p2turnl<3 && p2turnl>p2turnr) { p2av= ainc*1.5/LOOPSPTICK; } 
      if (p2turnr && p2turnr<3 && p2turnr>p2turnl) { p2av=-ainc*1.5/LOOPSPTICK; } 
    }
    else
    {
      if (p1turnl && p1turnl<3 && p1turnl>p1turnr) { p1av= ainc*1.5/smoothedloopcts[0]; } // av and p1turnl are zeroed on key up
      if (p1turnr && p1turnr<3 && p1turnr>p1turnl) { p1av=-ainc*1.5/smoothedloopcts[0]; } //  elsewhere, in the input section of the loop
      if (p2turnl && p2turnl<3 && p2turnl>p2turnr) { p2av= ainc*1.5/smoothedloopcts[0]; } 
      if (p2turnr && p2turnr<3 && p2turnr>p2turnl) { p2av=-ainc*1.5/smoothedloopcts[0]; } 
    }

    if(objdrawchngflag)
    {
      objdrawchngflag=0;
      while(rectilinearmode&&objdrawselect>2) objdrawselect=(objdrawselect+1)%OBJDRAWS;
      switch(objdrawselect)
      {
        case 0: drawobj=drawplayer; break;
        case 1: drawobj=drawplayernsr; break;
        case 2: drawobj=drawplayerring; break;
      }
    }

    if( startmixingprojs )
    {
      if( !intextscr ) if(nogrid) drawgridbo(0); else drawgrid(0);
      mixproja=1.0;
      mixprojold = projec ;
      do
      {
        if(startmixingprojs==2) projecti=(projecti+1)%PROJECTIONS;
        else projecti = rand()%PROJECTIONS ;
        setprojec(projecti);
      } while(projec==mixprojold);
      mixprojnew = projec ;
      projec = mixproj;
      //if(drawobj==drawplayern2) { drawobj=drawplayernsr; objdrawselect=1; }
      startmixingprojs = 0 ;
      mixingprojs = 1;
    }
    if( mixingprojs && !pausemixingprojs && !pause)
    {
      if( !intextscr ) if(nogrid) drawgridbo(0); else drawgrid(0);
      mixproja-=0.005;
      if(mixproja<=0.0)  // then done
      {
        projec = mixprojnew ;
        mixingprojs = pausemixingprojs = 0 ;
      }
      if( !intextscr ) if(nogrid) drawgridbo(1); else drawgrid(1);
    }

    if(startrectilinearmode)
    {
      //drawobj(p1long, p1lat, p1a, p1rad, 0, 0);
      //drawobj(p2long, p2lat, p2a, p2rad, 0, 0);
      if( !intextscr ) {  gridin(); clear_screen(); gtick=0; }
      if (startrectilinearmode==2) walledsquaresurface(); else squaresurface();
      projec = squareproj ;
      rectilinearmode=startrectilinearmode;
      startrectilinearmode=0;
    }
    if(changespace)
    {
      switch(changespace)
      {
        case 1: spheresurface(); break;
        case 2: squaresurface(); break;
        case 3: walledsquaresurface(); break;
      }
      changespace=0;
    }
    if(exitrectilinearmode)
    {
      //drawobj(p1long, p1lat, p1a, p1rad, 0, 0);
      //drawobj(p2long, p2lat, p2a, p2rad, 0, 0);
      if( !intextscr ) { gridout(); clear_screen(); gtick=0; } 
      spheresurface();
      setprojec(projecti);
      exitrectilinearmode=rectilinearmode=0;
    }

    if( autoclear && autocls(tock,autoclear) && !intextscr ) gtick=0;

    if ( !pause )
    {
      float dt=1.0, flt;  int dts=1, mtdts=1, h;
      flt=max(p1v,p2v);
      for(i=0;i<MAX_SHOTS;i++)
      {
        if(shots[i].owner!=-1) 
        {
          flt=max(flt,shots[i].v);
        }
      }
      if( flt>2.0*vinc ) { dts=ceil(flt/vinc); dt=1.0/dts; }
      if( movestiming ) 
      {
        flt = (float)LOOPSPTICK/prevloopcts; 
        mtdts=ceil(flt); dt*=flt/mtdts; 
      }
      //{ gotoxy(30,1); printf("%3d %3d",mtdts,dts); }
      for (h=0;h<mtdts;h++)
      for (i=0;i<dts;i++)
      {
        if(blackab.exists)
        {  
          float flt=0.0;
          blackabgravitar(dt);
          moveplayer(&(blackab.lon),&(blackab.lat),blackab.v,&(blackab.at),&flt,flt,dt);
        }
        moveplayer(&p1long,&p1lat,p1v,&p1at,&p1a,p1av,dt);
        moveplayer(&p2long,&p2lat,p2v,&p2at,&p2a,p2av,dt);
        moveshots(dt);
        if(blackab.exists)blackabcollisions();
        if (kollisions && p1exists && p2exists) p1p2collisions(dt);
        emergeshots(); //mark shots as having left their barrel
        if (shotplrkollisions) shotplrcollisions(shotkollangular,dt);
        if (moveplayer==walledsquaremoveplayer) { p1p2wallcollisions(); shotwallcollisions(); }
      }
      if (friction!=0) somefriction(friction-1,betterfriction,dt*dts*mtdts);
    }

    if( intextscr && ( tock>=gtick+1 || pause ) ) // update text screen
    {
      gtick = clock();
      if (tock>=keytick+866) pause = 1;
      switch( intextscr )
      {
        case 1:
          helpandstat(projecti,delnum,objdrawselect,autoclear,friction,p2pfriction,shotkollangular,erasetrails,eraseshottrails,
                      mixingprojs,rectilinearmode,physdisp,wallspin,betterfriction);
          break;
        case 2:
          morehelp( nogrid, logrid, kollisions, shotplrkollisions, rectilinearmode, physdisp, movestiming );
          break;
      }
    } 
    if(physdisp) 
    { 
      int i,textx=0;
      float plke,plpx,plpy, p1m,p2m,p1i,p2i;
      float tke,tpx,tpy,tlz00;
      vector ol000;
      static unsigned long int physdisploopct=0;
      if( loopctr-physdisploopct>prevloopcts || pause) 
      {
      physdisploopct = loopctr;
      p1m = p1rad*M_PI/Y_RES; p2m = p2rad*M_PI/Y_RES; 
      p1i = playermoment(p1m); p2i = playermoment(p2m);
      if(rectilinearmode)
      {
        float p1vx,p1vy,p2vx,p2vy;
        plke = 0.5*p1m*p1v*p1v + 0.5*p2m*p2v*p2v ;
        p1vx=p1v*cos(p1at); p2vx=p2v*cos(p2at); 
        p1vy=p1v*sin(p1at); p2vy=p2v*sin(p2at);
        plpx = p1m*p1vx + p2m*p2vx ;
        plpy = p1m*p1vy + p2m*p2vy ; 
        tpx = plpx; tpy = plpy; 
        tlz00 = p1m*(p1long*p1vy-p1lat*p1vx) + p1i*p1av + p2m*(p2long*p2vy-p2lat*p2vx) + p2i*p2av;
        tke = plke; 
        if(blackab.exists)
        {
          float bavx,bavy,bm;
          bm = blackab.m*M_PI/Y_RES;
          bavx = blackab.v * cos(blackab.at);
          bavy = blackab.v * sin(blackab.at);
          tpx += bm*bavx;
          tpy += bm*bavy;
          tlz00 += blackab.am + bm*(blackab.lon*bavy-blackab.lat*bavx);
          tke += 0.5*bm*blackab.v*blackab.v;
        }
      }
      else
      {
        float p1iorb,p2iorb;
        p1iorb = playerorbitmomentsphere(p1m);
        p2iorb = playerorbitmomentsphere(p2m);
        ol000 = vectsum( vecttimes( p1iorb, rcrossv(p1long,p1lat,p1v,p1at) ) , vecttimes( p2iorb, rcrossv(p2long,p2lat,p2v,p2at) ) ) ;
        plke = 0.5*p1iorb*p1v*p1v + 0.5*p2iorb*p2v*p2v ;      
        tke = plke; 
        if(blackab.exists)
        {
          float bm=blackab.m*M_PI/Y_RES;
          ol000 = vectsum(ol000, vecttimes(bm,rcrossv(blackab.lon,blackab.lat,blackab.v,blackab.at) ) );
          tke += 0.5*bm*blackab.v*blackab.v;
        }
      }
      for(i=0;i<MAX_SHOTS;i++)
      {
        if(shots[i].owner!=-1) 
        {
          float sv,sa,sm,slon,slat;
          sv = shots[i].v ; sa = shots[i].bear ; sm=0.5 * M_PI/Y_RES;
          slon=shots[i].lon; slat=shots[i].lat;
          if(rectilinearmode)
          {
            float svx,svy;
            svx = sv*cos(sa); svy = sv*sin(sa); 
            tpx += sm*svx; tpy += sm*svy;
            tlz00 += sm*(slon*svy-slat*svx) ;
          }
          else
          {
            vector shotl;
            shotl = vecttimes( sm, rcrossv(slon,slat,sv,sa) ) ;
            ol000.x += shotl.x; ol000.y += shotl.y; ol000.z += shotl.z; 
          }
          tke += 0.5*sm*sv*sv;  
        }
      }
      tke += p1i*p1av*p1av*0.5 + p2i*p2av*p2av*0.5;
      if(intextscr) textx=20;
      if(rectilinearmode) {
        gotoxy(1+textx,22); printf("plr linKE plyr Px   plyr Py   total Lz00"); 
        gotoxy(1+textx,23); printf("%+9.5f %+9.5f %+9.5f %+9.4f",plke*100000.0,plpx*1000.0,plpy*1000.0,tlz00*10000.0);
        gotoxy(1+textx,24); printf("totalKE   totalPx   totalPy");
        gotoxy(1+textx,25); printf("%+9.5f %+9.5f %+9.5f",tke*100000.0,tpx*1000.0,tpy*1000.0); }
      else {
        gotoxy(1+textx,22); printf("plr linKE"); if(intextscr)printf("                               ");
        gotoxy(1+textx,23); printf("%+9.5f",plke*100000.0); if(intextscr)printf("                               ");
        gotoxy(1+textx,24); printf("totalKE   orbital L xyz"); if(intextscr)printf("                  ");
        gotoxy(1+textx,25); printf("%+9.5f %+8.4f %+8.4f %+8.4f",tke*100000.0,ol000.x*10000.0,ol000.y*10000.0,ol000.z*10000.0); }
      if(intextscr) gotoxy(79,24);
      }
    }
    if ( blankphysdisp ) // clearphysdisp(&blankphysdisp,intextscr)
    {
      int textx=0;
      if(intextscr) textx=20;
      if(rectilinearmode) {
        gotoxy(1+textx,22); printf("                                        "); 
        gotoxy(1+textx,23); printf("                                        ");
        gotoxy(1+textx,24); printf("                                        ");
        gotoxy(1+textx,25); printf("                                       "); }
      else {
        gotoxy(1+textx,22); printf("         "); if(intextscr)printf("                               ");
        gotoxy(1+textx,23); printf("         "); if(intextscr)printf("                               ");
        gotoxy(1+textx,24); printf("                                        "); if(intextscr)printf("  ");
        gotoxy(1+textx,25); printf("                                       "); if(intextscr) printf(" "); }
      if(intextscr) gotoxy(79,24);
      blankphysdisp=0;
    }
    if(!intextscr)
    { if (((tock/2)*2)==tock) {
        if(p1exists)drawobj(p1long, p1lat, p1a, p1rad, p1colc, p1colf);
        if(p2exists)drawobj(p2long, p2lat, p2a, p2rad, p2colc, p2colf);
      }
      else {
        if(p2exists)drawobj(p2long, p2lat, p2a, p2rad, p2colc, p2colf);
        if(p1exists)drawobj(p1long, p1lat, p1a, p1rad, p1colc, p1colf);
      }
      drawshots();
      if(blackab.exists)
      {
        float dt=1.0, flt;  int mtdts;
        if( movestiming ) 
        {
          flt = (float)LOOPSPTICK/prevloopcts; 
          mtdts=ceil(flt); dt=flt/mtdts; 
        }
        drawblackab(dt);
      }
    }
    if( fpsreadout ) { gotoxy(2,1); printf("%4d",smoothedloopcts[0]*18); } //*CLOCKS_PER_SEC; //*18.2); }
    if(curioswitch2) {curiopauser(); keytick = clock();}    //
    //gotoxy(20,1); printf("%f %f",blackab.v,blackab.at); 
    //printf("%g %g",blackab.rad,blackab.rad2);
  }
  deinit_keyboard(); 

  tock = clock();
  if(!intextscr) vga_mode(0x03);
  printf("Time elapsed: %.1f seconds.\n", (tock - start) / CLK_TCK);
  printf("In 1/18.2 second clock ticks: %d.\n", (tock - start) );
  oldpauser();
  return 0;
}